package gov.va.cpe.vpr;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonView;
import gov.va.cpe.vpr.pom.AbstractPatientObject;
import gov.va.cpe.vpr.pom.IPatientObject;
import gov.va.cpe.vpr.pom.JSONViews;
import gov.va.hmp.healthtime.PointInTime;
import gov.va.hmp.vista.util.VistaStringUtils;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;

public class Problem extends AbstractPatientObject implements IPatientObject {

    private static final String PROBLEM = "Problem";

    private String localId;
    private String predecessor;
    private String successor;
    private String facilityCode;
    private String facilityName;
    private String locationCode;
    private String locationName;
    private String locationDisplayName;
    private String service;
    private String providerName;
    private String providerDisplayName;
    private String providerCode;
    private String problemType;
    private String problemText;
    private String code; // what code does this refer to?
    private String icdCode;
    private String icdName;
    private String statusCode;
    private String statusName;
    private String statusDisplayName;
    private String acuityCode;
    private String acuityName;
    private String history;
    private Boolean unverified;
    private Boolean removed;
    private PointInTime entered;
    private PointInTime updated;
    private PointInTime onset;
    private PointInTime resolved;
    private Boolean serviceConnected;
    private List<ProblemComment> comments;

    public Problem() {
        super(null);
    }

    @JsonCreator
    public Problem(Map<String, Object> vals) {
        super(vals);
    }

    @JsonView({JSONViews.JDBView.class, JSONViews.WSView.class, JSONViews.EventView.class})
    public String getLocalId() {
        return localId;
    }

    @JsonView({JSONViews.JDBView.class, JSONViews.WSView.class, JSONViews.EventView.class})
    public String getPredecessor() {
        return predecessor;
    }

    @JsonView({JSONViews.JDBView.class, JSONViews.WSView.class, JSONViews.EventView.class})
    public String getSuccessor() {
        return successor;
    }

    @JsonView({JSONViews.JDBView.class, JSONViews.WSView.class, JSONViews.EventView.class})
    public String getLocationCode() {
		return locationCode;
	}

	public String getLocationName() {
		return locationName;
	}

	public String getService() {
        return service;
    }
    @JsonView({JSONViews.JDBView.class, JSONViews.WSView.class, JSONViews.EventView.class})
	public String getProblemType() {
        return problemType;
    }

    public String getProviderName() {
		return providerName;
	}

    // not sure how to alias this in Solr so that it is called "provider_name" without colliding with getter
    @JsonView({JSONViews.JDBView.class, JSONViews.WSView.class, JSONViews.EventView.class})
    public String getProviderDisplayName() {
        if (providerDisplayName == null) {
            providerDisplayName = VistaStringUtils.nameCase(getProviderName());
        }
        return providerDisplayName;
    }

    @JsonView({JSONViews.JDBView.class, JSONViews.WSView.class, JSONViews.EventView.class})
    public String getProviderCode() {
		return providerCode;
	}

    public String getProblemText() {
        return problemText;
    }

    // not sure how to alias this in Solr so that it is called "location_name" without colliding with getter
    @JsonView({JSONViews.JDBView.class, JSONViews.WSView.class, JSONViews.EventView.class})
    public String getLocationDisplayName() {
        if (locationDisplayName == null) {
            locationDisplayName = VistaStringUtils.nameCase(getLocationName());
        }
        return locationDisplayName;
    }

    @JsonView({JSONViews.JDBView.class, JSONViews.WSView.class, JSONViews.EventView.class})
    public String getCode() {
        return code;
    }

    @JsonView({JSONViews.JDBView.class, JSONViews.WSView.class, JSONViews.EventView.class})
    public String getStatusCode() {
        return statusCode;
    }

    @JsonView({JSONViews.JDBView.class, JSONViews.WSView.class, JSONViews.EventView.class})
    public String getStatusName() {
        return statusName;
    }

    @JsonView({JSONViews.JDBView.class, JSONViews.WSView.class, JSONViews.EventView.class})
    public String getStatusDisplayName() {
        if (statusDisplayName == null) {
            statusDisplayName = VistaStringUtils.nameCase(getStatusName());
        }
        return statusDisplayName;
    }

    /**
     * Solr alias for 'statusDisplayName'.
     * @see #getStatusDisplayName()
     */
    @JsonView(JSONViews.SolrView.class)
    public String getProblemStatus() {
        return getStatusDisplayName();
    }

    public String getAcuityCode() {
		return acuityCode;
	}

	public String getAcuityName() {
		return acuityName;
	}

    @JsonView({JSONViews.JDBView.class, JSONViews.WSView.class, JSONViews.EventView.class})
	public String getHistory() {
        return history;
    }

    @JsonView({JSONViews.JDBView.class, JSONViews.WSView.class, JSONViews.EventView.class})
    public Boolean getUnverified() {
        return unverified;
    }

    public Boolean getRemoved() {
        return removed;
    }

    public PointInTime getEntered() {
        return entered;
    }

    public PointInTime getUpdated() {
        return updated;
    }

    public PointInTime getOnset() {
        return onset;
    }

    public PointInTime getResolved() {
        return resolved;
    }

    @JsonView({JSONViews.JDBView.class, JSONViews.WSView.class, JSONViews.EventView.class})
    public Boolean getServiceConnected() {
    	return (serviceConnected == null)?Boolean.FALSE:serviceConnected;
    }

    @JsonView({JSONViews.JDBView.class, JSONViews.WSView.class, JSONViews.EventView.class})
    public List<ProblemComment> getComments() {
        return comments;
    }

    /** Consolidated comment record, only used by SOLR */
    @JsonView(JSONViews.SolrView.class)
    public List<String> getComment() {
        if (comments == null || comments.isEmpty()) return Collections.emptyList();
        List<String> commentStrings = new ArrayList<>(comments.size());
        for (ProblemComment comment : this.getComments()) {
            commentStrings.add(comment.getComment());
        }
        return Collections.unmodifiableList(commentStrings);
    }

    public String getFacilityCode() {
        return facilityCode;
    }

    public String getFacilityName() {
        return facilityName;
    }

    public String getKind() {
        // we could potentially move this kind of logic to a "KindService(s)" if that is less smelly
        return PROBLEM;
    }

    @Override
    public String getSummary() {
        return problemText;
    }

    public List getTaggers() {
//        if (uid)
//            return manualFlush { Tagger.findAllByUrl(uid) }
//        else
//            return []
        return null;
        //TODO - fix this
    }

    public String getIcdCode() {
        return icdCode;
    }

    public String getIcdName() {
        return icdName;
    }
    
    /**
     * Get the ICD9 group code.
     * 
     * Example: "E819.9" => E819, 799.9 => 799
     */
    public String getIcdGroup() {
    	if (icdCode != null && icdCode.startsWith("urn:icd:") 
    			&& icdCode.length() >= 11 && icdCode.indexOf('.') > 0) {
    		return icdCode.substring(8, icdCode.indexOf('.'));
    	}
    	return null;
    }
}
