package gov.va.cpe.vpr.pom;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.annotation.JsonView;
import gov.va.cpe.vpr.UidUtils;

import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

public abstract class AbstractPatientObject extends AbstractPOMObject implements IPatientObject {
	protected List<POMIndex<?>> indexes; // extracted indexes from annotations
	protected String pid;
	protected WeakReference<IGenericPOMObjectDAO> daoRef; // hold reference to dao for potential lazy loading
	
	@JsonCreator
	public AbstractPatientObject(Map<String, Object> vals) {
		super(vals);
		
		// Initialize the data and clear events, since objects are not considered modified
		// unless updates are made after the constructor
		if (vals != null) {
			validate();
			clearEvents();
		}
	}
	
	// global getters ------------------------
	
	/**
	 * All domain objects have a patient identifier
	 * Equivalent to getPatient().getUID();
	 */
	@JsonProperty("pid")
	public String getPid() {
		return pid;
	}

    @JsonView(JSONViews.SolrView.class)
    public String getDomain() {
        return UidUtils.getCollectionNameFromUid(getUid());
    }

	// indexing ------------------------------
	
	/**
	 * Based on the declared index meta data, construct a list of indexed values.
	 * Subclasses can override or append to this list as needed.
	 */
	@JsonView(JSONViews.JDBView.class) // only include this node when serializing to disk
	@JsonIgnore // Ignore for now, not being used.
	public List<Map<String, Object>> getIDX() {
		List<Map<String, Object>> ret = new ArrayList<Map<String, Object>>();
		if (indexes == null) {
			// lazy-load index meta data
			indexes = POMIndex.extractIndexes(this.getClass());
		}
		
		for (POMIndex<?> idx : indexes) {
			idx.calcValue(this, ret);
		}
		return ret;
	}
	
	// life cycle ----------------------------
	protected void validate() {
		// values for UID and PID are required
//		if (uid == null || pid == null) {
//			// TODO: Is there a better exception class to use? Or make our own?
//			throw new IllegalArgumentException("UID and PID are required");
//		}
	}
	
	@Override
	public void load(IGenericPOMObjectDAO dao) {
		daoRef = new WeakReference<IGenericPOMObjectDAO>(dao);
	}
	
	@Override
	public void save(IGenericPOMObjectDAO dao) {
		// NOOP
	}

	@JsonIgnore
	public boolean isModified() {
		return modified;
	}

	@Override
    public String toString() {
        return getClass().getSimpleName() + "{" +
                "pid='" + pid + '\'' +
                ",uid='" + uid + '\'' +
                '}';
    }
}
