package gov.va.cpe.vpr.queryeng;

import com.google.common.base.Function;
import com.google.common.collect.Lists;
import gov.va.cpe.vpr.UidUtils;
import gov.va.cpe.vpr.queryeng.query.JDSQuery;
import gov.va.cpe.vpr.queryeng.query.QueryDef;
import org.apache.solr.client.solrj.SolrQuery;
import org.apache.solr.client.solrj.response.FacetField;
import org.apache.solr.client.solrj.response.QueryResponse;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;
import org.springframework.util.StringUtils;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * {@link ViewDef} for browsing through all patients in the VPR
 */
@Component(value = "gov.va.cpe.vpr.queryeng.VprPatientsViewDef")
@Scope("prototype")
public class VprPatientsViewDef extends ViewDef {
    public VprPatientsViewDef() {
        // fetch VPR PIDs for all patients in the VPR
        QueryDef qry = new QueryDef("syncstatus");
        qry.setForPatientObject(false);
        qry.where("forOperational").is("false");
        
        JDSQuery query = new JDSQuery("uid", qry);
        addQuery(query);

        // fetch domain counts per patient
        addQuery(new QueryMapper.PerRowAppendMapper(new JDSQuery("pid", new QueryDef(), "/vpr/{pid}/count/collection") {
            @Override
            protected void filterTransformResults(RenderTask task, Map<String, Object> params, List<Map<String, Object>> items) {
                RenderTask.RowRenderSubTask rowtask = (RenderTask.RowRenderSubTask) task;
                String pid = (String) rowtask.getParentRowVal("pid");
                Map<String, Integer> allVistaStats = getAllStatsFromSyncStatiiResult(rowtask.getParentRow());
                Map<String, Object> vistaKeyedStats = new HashMap<>();
                for(String key: allVistaStats.keySet()) {
                    // We can't compare raw VISTA to raw SOLR; SOLR has the converted domains such as lab->result and appt/visit->encounter.
                    // So, let's drive this off of our "official" UID-to-class map in UidUtils.
                    String domainKey = UidUtils.getDomainClass(key).getSimpleName().toLowerCase();
                    domainKey = domainKey.equals("medication")?"med":domainKey.equals("observation")?"obs":domainKey;

                    Integer vval = (Integer)vistaKeyedStats.get(getVistaDomainCountKey(domainKey));
                    if(vval!=null) {
                        vval = vval + allVistaStats.get(key);
                    }
                    else {
                        vval = allVistaStats.get(key);
                    }
                    // Put the total JDS count in the row
                    vistaKeyedStats.put(getVistaDomainCountKey(domainKey), vval);
                }
                task.appendRow(pid, vistaKeyedStats);
                Map<String, Object> jdsKeyedStats = new HashMap<>();
                for(Map<String, Object> item: items) {
                    String key = (String)item.get("topic");
                    String domainKey = UidUtils.getDomainClass(key).getSimpleName().toLowerCase();
                    //domainKey = domainKey.equals("medication")?"med":domainKey.equals("observation")?"obs":domainKey;
                    Integer jval = (Integer)jdsKeyedStats.get(getJdsDomainCountKey(domainKey));
                    if(jval==null) {
                        jval = (Integer)item.get("count");
                    } else {
                        jval = jval + (Integer)item.get("count");
                    }
                    jdsKeyedStats.put(getJdsDomainCountKey(domainKey), jval);
                }
                task.appendRow(pid, jdsKeyedStats);
            }
        }));
        // fetch domain counts per patient from Solr
        addQuery(new QueryMapper.PerRowAppendMapper(new SOLRFacetQuery("pid", "pid:#{getParentRowVal('pid')}", "domain") {
            @Override
            protected void modifySolrQuery(SolrQuery qry, RenderTask task) {
                super.modifySolrQuery(qry, task);
                qry.setFacetMinCount(0);
            }

            @Override
            protected void mapResults(RenderTask task, QueryResponse resp) {
                FacetField ff = resp.getFacetFields().get(0);
                if (ff == null || ff.getValues() == null) {
                    return;
                }

                if (task instanceof RenderTask.RowRenderSubTask) {
                    Map<String, Object> map = new HashMap<String, Object>();
                    map.put(getPK(), ((RenderTask.RowRenderSubTask) task).getParentRowVal(getPK()));
                    for (FacetField.Count c : ff.getValues()) {
                        String key = c.getName();
                        if ( key != null ) {
                            Class dCls = UidUtils.getDomainClass(key);  // use this convert pov and cpt received from solr to purposeofvisit and visitcptcode
                            if ( dCls != null ) {
                                String domainKey = dCls.getSimpleName().toLowerCase();
                                if ( c.getCount() > 0 ) {
                                    map.put(getSolrDomainCountKey(domainKey), c.getCount());
                                }
                            }
                            else {  // we receive result, encounter (for appointment, visit), procedure (for procedure, surgery, consult, image) from solr
                                if ( c.getCount() > 0 ) {
                                    map.put(getSolrDomainCountKey(key), c.getCount());
                                }
                            }
                        }
                    }
                    task.add(mapRow(task, map));
                }
            }

            @Override
            protected Map<String, Object> mapRow(RenderTask renderer, Map<String, Object> row) {
                row.put("jdsSolrCountMismatch", isJdsSolrCountMismatch(row));
                return row;
            }

            private boolean isJdsSolrCountMismatch(Map<String, Object> row) {
                for (String domain : UidUtils.getAllPatientDataDomains()) {
                    String jdsKey = getJdsDomainCountKey(domain);
                    String solrKey = getSolrDomainCountKey(domain);
                    Object jdsValue = row.get(jdsKey);
                    Object solrValue = row.get(solrKey);
                    if (jdsValue == null && solrValue == null) continue;
                    if ((jdsValue != null && solrValue == null) || (jdsValue == null && solrValue != null)) {
                        return true;
                    }
                    if (!jdsValue.equals(solrValue)) {
                        return true;
                    }
                    return false;
                }
                return false;
            }
        }));

        addColumns(query,
                "pid",
                "displayName",
                "error");

        List<String> domains = new ArrayList(UidUtils.getAllPatientDataDomainClasses());
        List<String> vistaDomains = Lists.transform(domains, new Function<String, String>() {
            public String apply(String input) {
                return getVistaDomainCountKey(input);
            }
        });
        List<String> jdsDomains = Lists.transform(domains, new Function<String, String>() {
            public String apply(String input) {
                return getJdsDomainCountKey(input);
            }
        });
        List<String> solrDomains = Lists.transform(domains, new Function<String, String>() {
            public String apply(String input) {
                return getSolrDomainCountKey(input);
            }
        });

        addColumns(query, jdsDomains);
        addColumns(query, solrDomains);
    }

    private Map<String, Integer> getAllStatsFromSyncStatiiResult(Map<String, Object> parentRow) {
        Map<String, Integer> allDomainTotals = new HashMap<>();
        Object stat = parentRow.get("syncStatusByVistaSystemId");
        if((stat != null) && (stat instanceof Map)) {
            for(Object systemMap: ((Map)stat).values()) {
                Object dtots = ((Map)systemMap).get("domainExpectedTotals");
                if(dtots!=null && dtots instanceof Map) {
                    for(Object key: ((Map)dtots).keySet()) {
                        Object countsMap = ((Map)dtots).get(key);
                        if(countsMap!=null && countsMap instanceof Map) {
                            Object count = ((Map)countsMap).get("count");
                            if(count!=null) {
                                Integer cnt = Integer.parseInt(count.toString());
                                allDomainTotals.put(key.toString(), allDomainTotals.get(key.toString())!=null?allDomainTotals.get(key.toString()):0 + cnt);
                            }
                        }
                    }
                }
            }
        }
        return allDomainTotals;
    }

    private String getVistaDomainCountKey(String key) {
        return "vista" + StringUtils.capitalize(key);
    }

    private String getJdsDomainCountKey(String key) {
        return "jds" + StringUtils.capitalize(key);
    }

    private String getSolrDomainCountKey(String key) {
        return "solr" + StringUtils.capitalize(key);
    }
}
