package gov.va.cpe.vpr.sync.msg;

import gov.va.cpe.vpr.Encounter;
import gov.va.cpe.vpr.UidUtils;
import gov.va.cpe.vpr.dao.ISolrDao;
import gov.va.cpe.vpr.dao.IVprSyncErrorDao;
import gov.va.cpe.vpr.pom.IGenericPatientObjectDAO;
import gov.va.cpe.vpr.pom.IPatientObject;
import gov.va.cpe.vpr.pom.jds.JdsOperations;
import gov.va.cpe.vpr.sync.SyncError;
import gov.va.cpe.vpr.sync.SyncMessageConstants;
import gov.va.hmp.HmpProperties;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.EnvironmentAware;
import org.springframework.core.env.Environment;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.jms.listener.SessionAwareMessageListener;
import org.springframework.jms.support.converter.SimpleMessageConverter;
import org.springframework.util.Assert;

import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.Session;

import java.util.Map;
import java.util.Set;

public class ReindexPatientMessageHandler implements SessionAwareMessageListener, EnvironmentAware {

	private IVprSyncErrorDao errorDao;

    @Autowired
    public void setErrorDao(IVprSyncErrorDao errorDao) {
        this.errorDao = errorDao;
    }

    private static Logger LOGGER = LoggerFactory.getLogger(ReindexPatientMessageHandler.class);

    private ISolrDao solrDao;

    private IGenericPatientObjectDAO genericPatientRelatedDao;

	private Environment environment;

    public ISolrDao getSolrDao() {
        return solrDao;
    }

    @Autowired
    public void setSolrDao(ISolrDao solrDao) {
        this.solrDao = solrDao;
    }

    @Override
	public void setEnvironment(Environment environment) {
		this.environment = environment;
	}

	public IGenericPatientObjectDAO getGenericPatientRelatedDao() {
        return genericPatientRelatedDao;
    }

    @Autowired
    public void setGenericPatientRelatedDao(IGenericPatientObjectDAO genericPatientRelatedDao) {
        this.genericPatientRelatedDao = genericPatientRelatedDao;
    }

    @Autowired
    SimpleMessageConverter converter;

    public void onMessage(Message message, Session session) {
    	Integer batchSize = this.environment.getProperty(HmpProperties.SOLR_REINDEX_BATCH_SIZE, Integer.class, new Integer(500));
        Map msg = null;
        try {
            msg = (Map) converter.fromMessage(message);
            final String pid = (String) msg.get(SyncMessageConstants.PATIENT_ID);
            Assert.hasText(pid, "[Assertion failed] - the 'pid' argument must have text; it must not be null, empty, or blank");

            LOGGER.debug("Reindexing " + pid);

            Map<String, Long> solrCounts = solrDao.findDomainCounts(pid); // key - simpleName of domain class
            Map<String, Long> jdsCounts = genericPatientRelatedDao.findDomainCountsByPid(pid);   // key - simpleName of domain class

            for (String domain : getAllPatientDataDomainNames()) {
                Class domainClass = getDomainClass(domain);

                String simpleDomainName = domainClass.getSimpleName().toLowerCase();
                Long jdsDomainCnt = jdsCounts.get(simpleDomainName);
                Long solrDomainCnt = solrCounts.get(simpleDomainName);

                boolean reindex = false;

                if ( jdsDomainCnt == null || jdsDomainCnt == 0 ) {
                    if ( solrDomainCnt != null && solrDomainCnt > 0) {
                        // delete this domain from solr
                        solrDao.deleteByQuery("pid:" + pid + " AND domain:" + simpleDomainName);
                    }
                }
                else { // jdsDomainCnt > 0
                    if ( solrDomainCnt == null || solrDomainCnt == 0 ) {
                        reindex = true;
                    }
                    else if ( solrDomainCnt.longValue() != jdsDomainCnt.longValue() ) {
                        // delete domain in solr
                        solrDao.deleteByQuery("pid:" + pid + " AND domain:" + simpleDomainName);
                        reindex = true;
                    }
                }

                if ( reindex ) {
                    LOGGER.debug("*** Re-indexing for PID = " + pid + "  Domain = " + simpleDomainName + "  jdsCnt = " + jdsDomainCnt + "  solrCnt = " + solrDomainCnt);

                    // re-index in chunks to prevent jams
                    PageRequest page = new PageRequest(0, batchSize);
                    Page<IPatientObject> items = genericPatientRelatedDao.findAllByPID(domainClass, pid, page);
                    while (items != null) {

                        // reindex each item in the page/chunk
                        for (IPatientObject item : items) {
                            try {
                                solrDao.save(item);
                            } catch (Exception e) {
                                LOGGER.error("unable to reindex item " + String.valueOf(item), e);
                        }
                        }

                        // fetch next page (if any), being very careful to avoid infinite loop
                        // potential of a mis-configured page object
                        if (items.getSize() > 0 && !items.isLastPage()) {
                            page = new PageRequest(page.getPageNumber() + 1, page.getPageSize());
                            items = genericPatientRelatedDao.findAllByPID(domainClass, pid, page);
                        } else {
                            // done, exit
                            items = null;
                            break;
                        }
                    }
                }
            }
        } catch (Exception e) {
            LOGGER.error("Error synching patient data: " + e.getMessage(), e);
            try {
                errorDao.save(new SyncError(message, msg, e));
                session.recover();
            } catch (JMSException e1) {
                LOGGER.error("Exception saving sync error: " + e.getMessage(), e1);
            }
        }
    }

    // to avoid multiple times of solr re-index for procedure and encounter
    private Set<String> getAllPatientDataDomainNames() {

        Set<String> ptDataDomains = UidUtils.getAllPatientDataDomains();

        ptDataDomains.remove("image");
        ptDataDomains.remove("surgery");
        ptDataDomains.remove("consult");
        ptDataDomains.remove("appointment");
        ptDataDomains.remove("visit");

        ptDataDomains.add("encounter");

        return ptDataDomains;
    }

    private Class getDomainClass(String domain) {

        if ( domain.equalsIgnoreCase("encounter") ) {
            return Encounter.class;
        }

        return UidUtils.getDomainClass(domain);
    }
}
