
package gov.va.cpe.vpr.sync.vista;

import com.codahale.metrics.MetricRegistry;
import com.codahale.metrics.Timer;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.node.ArrayNode;
import gov.va.cpe.vpr.IBroadcastService;
import gov.va.cpe.vpr.PatientDemographics;
import gov.va.cpe.vpr.PidUtils;
import gov.va.cpe.vpr.UidUtils;
import gov.va.cpe.vpr.dao.IVprSyncStatusDao;
import gov.va.cpe.vpr.dao.RoutingDataStore;
import gov.va.cpe.vpr.pom.IPatientDAO;
import gov.va.cpe.vpr.pom.POMUtils;
import gov.va.cpe.vpr.sync.ISyncService;
import gov.va.cpe.vpr.sync.SyncMessageConstants;
import gov.va.cpe.vpr.sync.SyncStatus;
import gov.va.hmp.HmpProperties;
import gov.va.hmp.healthtime.PointInTime;
import gov.va.hmp.hub.dao.IVistaAccountDao;
import gov.va.hmp.vista.rpc.JacksonRpcResponseExtractor;
import gov.va.hmp.vista.rpc.RpcOperations;
import gov.va.hmp.vista.rpc.RpcResponse;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.EnvironmentAware;
import org.springframework.core.convert.ConversionFailedException;
import org.springframework.core.convert.ConversionService;
import org.springframework.core.env.Environment;
import org.springframework.dao.DataRetrievalFailureException;
import org.springframework.stereotype.Component;
import org.springframework.util.StringUtils;

import java.util.*;

import static gov.va.cpe.vpr.sync.vista.SynchronizationRpcConstants.*;
import static gov.va.hmp.vista.util.RpcUriUtils.VISTA_RPC_BROKER_SCHEME;

@Component
public class VistaVprDataExtractEventStreamDAO implements IVistaVprDataExtractEventStreamDAO, EnvironmentAware {

    public static final String EXTRACT_SCHEMA = "3.001";

    static final int DEFAULT_BATCH_SIZE = 1000;
    static final int DEFAULT_MAX_SIZE_BYTES = 0;
    static final int DEFAULT_TARGET_TIME_SECONDS = 0;
    static final Logger LOG = LoggerFactory.getLogger(VistaVprDataExtractEventStreamDAO.class);

    private RpcOperations synchronizationRpcTemplate;
    private IPatientDAO patientDao;
    private RoutingDataStore routingDao;
    private MetricRegistry metricRegistry;
    private ConversionService conversionService;
    private IVprSyncStatusDao syncStatusDao;
    private IBroadcastService bcSvc;
    private ISyncService syncService;
    private Environment environment;

    private JacksonRpcResponseExtractor jsonExtractor = new JacksonRpcResponseExtractor();
    private int asyncBatchSize;
    private int asyncMaxSizeBytes;
    private int asyncTargetTimeSeconds;
    private Boolean hdrEnabled;
    private IVistaAccountDao vistaAccountDao;

    @Autowired
    public void setVistaAccountDao(IVistaAccountDao vistaAccountDao) {
        this.vistaAccountDao = vistaAccountDao;
    }

    @Autowired
    public void setBroadcastService(IBroadcastService bcSvc) {
        this.bcSvc = bcSvc;
    }

    @Autowired
    public void setSynchronizationRpcTemplate(RpcOperations synchronizationRpcTemplate) {
        this.synchronizationRpcTemplate = synchronizationRpcTemplate;
    }

    @Autowired
    public void setPatientDao(IPatientDAO patientDao) {
        this.patientDao = patientDao;
    }

    @Autowired
    public void setRoutingDao(RoutingDataStore routingDao) {this.routingDao = routingDao;}

    @Autowired
    public void setMetricRegistry(MetricRegistry metricRegistry) {
        this.metricRegistry = metricRegistry;
    }

    @Override
    public String fetchVprVersion(String vistaId) {
        return synchronizationRpcTemplate.executeForString(VISTA_RPC_BROKER_SCHEME + "://" + vistaId + VPR_DATA_VERSION_RPC_URI);
    }

    @Override
    public VistaDataChunk fetchPatientDemographicsWithDfn(String vistaId, String ptDfn) {
        return fetchPatientDemographics(vistaId, ptDfn, false);
    }

    @Override
    public VistaDataChunk fetchPatientDemographicsWithIcn(String vistaId, String ptIcn) {
        return fetchPatientDemographics(vistaId, ptIcn, true);
    }

    @Autowired
    public void setSyncService(ISyncService syncService) {
        this.syncService = syncService;
    }

    @Autowired
    public void setConversionService(ConversionService conversionService) {
        this.conversionService = conversionService;
    }

    @Autowired
    public void setSyncStatusDao(IVprSyncStatusDao syncStatusDao) {
        this.syncStatusDao = syncStatusDao;
    }

    private VistaDataChunk fetchPatientDemographics(String vistaId, String icnOrDfn, boolean isIcn) {
        Timer.Context timer = metricRegistry.timer(MetricRegistry.name("vpr.fetch.patient")).time();
        try {
            Map rpcArg = new HashMap();
            rpcArg.put("patientId", (isIcn ? ";" + icnOrDfn : icnOrDfn));
            rpcArg.put("domain", "patient");
            rpcArg.put("extractSchema",EXTRACT_SCHEMA);
            RpcResponse response = synchronizationRpcTemplate.execute(VISTA_RPC_BROKER_SCHEME + "://" + vistaId + VPR_GET_VISTA_DATA_JSON_RPC_URI, rpcArg);
            JsonNode json = jsonExtractor.extractData(response);
            JsonNode patientJsonNode = json.path("data").path("items").path(0);
            if (patientJsonNode.isNull())
                throw new DataRetrievalFailureException("missing 'data.items[0]' node in JSON RPC response");
            VistaDataChunk patientChunk = VistaDataChunk.createVistaDataChunk(vistaId, response.getRequestUri(), patientJsonNode, "patient", 0, 1, null, VistaDataChunk.getProcessorParams(vistaId, icnOrDfn, isIcn));
            patientChunk.getParams().put(SyncMessageConstants.DIVISION, response.getDivision());
            patientChunk.getParams().put(SyncMessageConstants.DIVISION_NAME, response.getDivisionName());

            if (!isIcn)
                patientChunk.setLocalPatientId(icnOrDfn);

            return patientChunk;
        } catch (RuntimeException e) {
            throw e;
        } finally {
            timer.stop();
        }
    }

    public VistaDataChunk fetchOneByUid(String vistaId, String pid, String uid) {
        Timer.Context timer = metricRegistry.timer(MetricRegistry.name("vpr.fetch.patient")).time();
        String domain = UidUtils.getCollectionNameFromUid(uid);
        PatientDemographics pt = patientDao.findByPid(pid);
        try {
            Map rpcArg = new HashMap();
            rpcArg.put("uid", uid);
            RpcResponse response = synchronizationRpcTemplate.execute(VISTA_RPC_BROKER_SCHEME + "://" + vistaId + VPR_GET_VISTA_DATA_JSON_RPC_URI, rpcArg);
            JsonNode json = jsonExtractor.extractData(response);
            JsonNode jsonNode = json.path("data").path("items").path(0);
            if (jsonNode.isNull())
                throw new DataRetrievalFailureException("missing 'data.items[0]' node in JSON RPC response");
            VistaDataChunk chunk = VistaDataChunk.createVistaDataChunk(vistaId, response.getRequestUri(), jsonNode, domain, 0, 1, pt);

            return chunk;
        } catch (RuntimeException e) {
            throw e;
        } finally {
            timer.stop();
        }
    }

    @Override
    public synchronized void unsubscribePatient(String vistaId, String pid) {
        Map<String, Object> params = new HashMap<>();
        params.put("hmpSrvId", environment.getProperty(HmpProperties.SERVER_ID));
        params.put("pid", pid);
        String url = VISTA_RPC_BROKER_SCHEME + "://" + vistaId + VPR_STREAM_DELSUB_RPC_URI;
        SyncStatus stat = syncStatusDao.findOneByPid(pid);
        synchronizationRpcTemplate.executeForJson(url, params);
        if (stat != null) {
            syncStatusDao.delete(stat);
        } else {
            LOG.warn("Removing pt subscription; No sync status object found for pid: " + pid);
        }
    }

    @Override
    public void unsubscribePatient(String vistaId, PatientDemographics pt) {
        unsubscribePatient(vistaId, pt.getPid());
    }

    @Override
    public synchronized void subscribePatient(String vistaId, String pid) {
        if(!vistaId.equalsIgnoreCase(vistaAccountDao.getPrimaryVistaSystemId())) {
            LOG.warn("putSubscription called with vistaID: "+vistaId+" and pid: "+pid+" but this does not match the primary vista ID: "+vistaAccountDao.getPrimaryVistaSystemId()+"; ; No synchronization for this patient will be triggered.");
            return;
        }
        Map<String, Object> params = new HashMap<>();
        params.put("server", environment.getProperty(HmpProperties.SERVER_ID));
        params.put("command", "putPtSubscription");

        String dfn = PidUtils.getDfn(pid);
        if (StringUtils.hasText(dfn)) {
            params.put("localId", dfn);
        } else {
            throw new IllegalArgumentException("Missing dfn for subscribing patients.");
        }

        SyncStatus stat = syncStatusDao.findOneByPid(pid);
        if (stat != null) {
            LOG.warn("putSubscription called but patient already subscribed; pid={}; No synchronization for this patient will be triggered.", pid);
            return;
        }

        String url = VISTA_RPC_BROKER_SCHEME + "://" + vistaId + VPR_STREAM_API_RPC_URI;
        JsonNode json = synchronizationRpcTemplate.executeForJson(url, params);

        if (json.has("error")) {
            String msg = json.path("error").path("message").asText();
            LOG.error("VistA event stream error: {}", msg);
        } else {
        	PatientDemographics pt = patientDao.findByPid(pid);
        	if(pt!=null) {
	            stat = createSyncStatus(vistaId, pt);
	            stat.setQueueStatus(vistaId, json);
	            syncStatusDao.save(stat);
        	}
        	else {
        		LOG.warn("putSubscription called for patient that does not exist: "+pid);
        		return;
        	}
        }
    }

    private SyncStatus createSyncStatus(String vistaId, PatientDemographics pt) {
        SyncStatus stat = new SyncStatus();
        stat.setData("pid", pt.getPid());
        stat.setData("displayName",pt.getDisplayName());
        
        Set<String> qdfns = new HashSet<String>(pt.getPatientIds());
        qdfns.remove(pt.getIcn());
        qdfns.remove(pt.getSsn());
        qdfns.remove(pt.getPid());
        stat.setData("localPatientIds", StringUtils.collectionToCommaDelimitedString(qdfns));

        SyncStatus.VistaAccountSyncStatus vstat = new SyncStatus.VistaAccountSyncStatus();
        vstat.setPatientUid(UidUtils.getPatientUid(vistaId, pt.getLocalPatientIdForSystem(vistaId)));
        vstat.setDfn(pt.getLocalPatientIdForSystem(vistaId));

        Map<String, Object> syncStatusByVistaSystemId = new HashMap<>();
        syncStatusByVistaSystemId.put(vistaId, vstat);

        stat.setData("syncStatusByVistaSystemId", syncStatusByVistaSystemId);
        return stat;
    }

    // It now serves both patient and operational data purposes.
    @Override
    public synchronized VprUpdateData fetchUpdates(String vistaId, String division, String previousLastUpdate) {
        // Synchronized because, in the case where there is a change to patient subscription,
        // either subscribing new or removing a patient subscription,
        // we don't want to collide with status checks.
        // We can safely move this when we move the sync status check elsewhere.


        Map<String, Object> params = new HashMap<>();

        params.put("command", "getPtUpdates");
        params.put("lastUpdate", previousLastUpdate);

        if(asyncMaxSizeBytes>0) {
            params.put("maxSize", asyncMaxSizeBytes);
        } else {
            params.put("max", asyncBatchSize);
        }

        String version = environment.getProperty(HmpProperties.VERSION);
        params.put("hmpVersion", version);
        params.put("extractSchema",EXTRACT_SCHEMA);

        List<VistaDataChunk> chunks = new ArrayList<VistaDataChunk>();

        params.put("server", environment.getProperty(HmpProperties.SERVER_ID));

        String url = VISTA_RPC_BROKER_SCHEME + "://" + vistaId + VPR_STREAM_API_RPC_URI;

        long reqStart = System.currentTimeMillis();
        JsonNode json = synchronizationRpcTemplate.executeForJson(url, params);
        long reqStop = System.currentTimeMillis();
        Integer remainingObjects = json.path("data").path("remainingObjects").asInt();

        if(remainingObjects>0 && asyncMaxSizeBytes>0 && asyncTargetTimeSeconds>0) {
            long dR = reqStop-reqStart;
            if(dR>0) {
                double dRatio = new Double(asyncTargetTimeSeconds * 1000) / dR;
                if (dRatio < 0.5) {
                    dRatio = 0.5;
                }
                if (dRatio > 2) {
                    dRatio = 2;
                }
                asyncMaxSizeBytes = new Double(asyncMaxSizeBytes * dRatio).intValue();
            }
        }

        JsonNode data = json.path("data");
        String newlyReceivedUpdate = data.get("lastUpdate").asText();

        ArrayNode items = (ArrayNode) json.findValue("items");

        Map<String, PatientDemographics> patMap = new HashMap<>();
        Map<String, SyncStatus> statMap = new HashMap<>();
        VprUpdateData rslt = new VprUpdateData();
        for (JsonNode item : items) {
            boolean isBatch = false;
            JsonNode bnode = item.path("batch");
            if(bnode!=null && !bnode.isMissingNode()) {
                isBatch = bnode.asBoolean();
            }
            String domainName = item.path("collection").asText();
            int tot = item.path("total").asInt();
            int idx = item.path("seq").asInt();
            String pid = item.path("pid").asText();
            JsonNode object = item.path("object");
            if(domainName.equalsIgnoreCase("syncCommand")) {
                handleSyncCommand(vistaId, chunks, url, item, object);
            } else if (domainName.equalsIgnoreCase("patient") || !StringUtils.hasText(pid)) {   // If the record is basically operational data (not tied to an existing patient)
                if (!item.path("deletes").isMissingNode()) {
                    JsonNode deletesNode = item.path("deletes");
                    for (JsonNode deleteNode : deletesNode) {
                        String uid = deleteNode.get("uid").textValue();
                        rslt.getUidsToDelete().add(uid);
                    }
                } else if (!item.path("error").isMissingNode()) {
                    JsonNode errorsNode = item.path("error");
                    for (JsonNode errorNode : errorsNode) {
                        String msg = errorNode.path("error").textValue();
                        LOG.error("error in VPR update data: " + msg);
                    }
                } else {
                    handleOperationalDataItem(vistaId, chunks, url, domainName, tot, idx, object, statMap, isBatch);
                }

            } else {
                handlePatientDataItem(vistaId, division, chunks, url, patMap, item, domainName, tot, idx, pid, object, statMap, isBatch);
            }
        }

        ArrayNode wpd = (ArrayNode)data.path("waitingPids");
        List<String> pendingPids = new ArrayList<>();
        for(JsonNode wp: wpd) {pendingPids.add(wp.asText());}

        ArrayNode ppd = (ArrayNode)data.path("processingPids");
        List<String> processingPids = new ArrayList<>();
        for(JsonNode pp: ppd) {processingPids.add(pp.asText());}

        for(String pid: pendingPids) {
            SyncStatus stat = statMap.get(pid);
            if(stat==null) {
                stat = syncStatusDao.findOneByPid(pid);
            }
            if(stat!=null) {
                stat.getSyncStatusByVistaSystemId().get(vistaId).setQueuePosition(pendingPids.indexOf(pid)+1);
                statMap.put(pid, stat);
            }
        }

        for(String pid: processingPids) {
            SyncStatus stat = statMap.get(pid);
            if(stat==null) {
                stat = syncStatusDao.findOneByPid(pid);
            }
            if(stat!=null) {
                stat.getSyncStatusByVistaSystemId().get(vistaId).setProcessingPosition(processingPids.indexOf(pid)+1);
                statMap.put(pid, stat);
            }
        }

        for (SyncStatus stat : statMap.values()) {
            syncStatusDao.save(stat);
            broadcastSyncStatus(stat);
        }

        rslt.setChunks(chunks);
        rslt.setLastUpdate(newlyReceivedUpdate);
        return rslt;
    }

    public JsonNode getVistaHealthCheck(String vistaId, String division) {
        Map<String, Object> params = new HashMap<>();
        params.put("command", "checkHealth");
        params.put("server", environment.getProperty(HmpProperties.SERVER_ID));
        String url = VISTA_RPC_BROKER_SCHEME + "://" + vistaId + VPR_STREAM_API_RPC_URI;
        return synchronizationRpcTemplate.executeForJson(url, params);
    }

    private void handleSyncCommand(String vistaId, List<VistaDataChunk> chunks, String url, JsonNode item, JsonNode object) {
        String command = object.path("command").textValue();
        if(command.equalsIgnoreCase("deleteDomainForPatient") || command.equalsIgnoreCase("deleteUidForPatient")) {
            String pid = object.path("pid").textValue();
            PatientDemographics dem = patientDao.findByPid(pid);
            if(dem!=null) {
                VistaDataChunk chunk = VistaDataChunk.createVistaDataChunk(vistaId, url, object, "syncCommand", 0, 0, dem, VistaDataChunk.getProcessorParams(vistaId, dem.getIcn(), true), true, VistaDataChunk.COMMAND);
                chunks.add(chunk);
            } else {
                // The patient was removed, so the domain is already gone. We can safely ignore this command message for this patient.
                LOG.info("Patient "+pid+" was missing when deleteDomainForPatient command message was received.");
            }
        } else if(command.equalsIgnoreCase("deleteOperationalDomain") || command.equalsIgnoreCase("deleteOperationalUid")) {
            VistaDataChunk chunk = VistaDataChunk.createVistaDataChunk(vistaId, url, object, "syncCommand", 0, 0, null, VistaDataChunk.getProcessorParams(vistaId, null, false), true, VistaDataChunk.COMMAND);
            chunks.add(chunk);
        }
    }


    private void broadcastSyncStatus(SyncStatus stat) {
        Map<String, Object> message = new HashMap<String, Object>();
        message.put("eventName", "syncStatusChange");
        message.put("syncStatus", stat.getData());
        bcSvc.broadcastMessage(message);
    }

    private void handlePatientDataItem(String vistaId, String division, List<VistaDataChunk> chunks,
                                       String url, Map<String, PatientDemographics> patMap, JsonNode item,
                                       String domainName, int tot, int idx, String pid, JsonNode object, Map<String, SyncStatus> statMap, boolean isBatch) {
        if (domainName.equals("syncStatus")) {
            // Now, update sync status stuff.
            // This should be the end of this patient's sync stream - so if we have a mishmash, we may want to acknowledge it as complete with errors.
            PatientDemographics pt = patientDao.findByPid(pid);
            if (pt != null) {
                SyncStatus stat = statMap.get(pt.getPid());
                if (stat == null) {
                    stat = syncStatusDao.findOneByPid(pt.getPid());
                }
                if (stat == null) {
                    String dfn = pt.getLocalPatientIdForSystem(vistaId);
                    stat = createSyncStatus(vistaId, pt);
                }
                stat.setSyncComplete(vistaId, true);

                JsonNode dataNode = POMUtils.parseJSONtoNode(stat.toJSON());
                VistaDataChunk chunk = VistaDataChunk.createVistaDataChunk(vistaId, url, dataNode, "syncStatus", idx, tot, pt, VistaDataChunk.getProcessorParams(vistaId, pid, false), false);
                chunks.add(chunk);//syncService.sendImportVistaDataExtractItemMsg(chunk);
                //statMap.put(pt.getPid(), stat);
                statMap.remove(pt.getPid());
                pt.setLastUpdated(PointInTime.now());
                routingDao.save(pt);

                if(this.hdrEnabled) {
                    // TODO: How do I get the correct division?
                   syncService.sendHdrPatientImportMsg(pid, division, vistaId);
                }
            }

        } else if (domainName.equals("syncStart")) {
            // Handle initial patient demographic object saving here
            if (object.isNull())
                throw new DataRetrievalFailureException("missing 'data.items[0]' node in JSON RPC response");
            VistaDataChunk patientChunk = VistaDataChunk.createVistaDataChunk(vistaId, url, object, "patient", 0, 1, null, VistaDataChunk.getProcessorParams(vistaId, pid, false), false);

            try {
                PatientDemographics pt = conversionService.convert(patientChunk, PatientDemographics.class);
                pt = routingDao.save(pt);

                String localPatientId = pt.getLocalPatientIdForSystem(vistaId);
                patientChunk.setLocalPatientId(localPatientId);
                SyncStatus stat = statMap.get(pt.getPid());
                if (stat == null) {
                    stat = syncStatusDao.findOneByPid(pt.getPid());
                }
                if (stat == null) {
                    stat = createSyncStatus(vistaId, pt);
                }
                statMap.put(pt.getPid(), stat);
            } catch(ConversionFailedException e){
                LOG.error("Unable to convert patientChunk into PatientDemographics object: " + patientChunk, e);
            }
        } else if (tot > 0 && idx > 0) {
            PatientDemographics pt = patMap.get(pid);
            if (pt == null) {
                pt = patientDao.findByPid(pid);
                if (pt == null && item.path("pid") != null) {
                    LOG.warn("JDS should have a record for pid: " + pid + " but it cannot be found; Requesting patient record from VISTA...");
                    VistaDataChunk patientChunk = this.fetchPatientDemographics(vistaId, PidUtils.getDfn(pid), false);
                    pt = conversionService.convert(patientChunk, PatientDemographics.class);
                    if (pt == null) {
                        LOG.error("Cannot find patient by pid: " + pid + " Item data will be skipped for uid: " + item.path("uid").asText(), new DataRetrievalFailureException("Can't find patient by pid: " + pid + " when checking async stream"));
                        return;
                    }
                    LOG.warn("VISTA record for pid: " + pid + " has been recovered and saved to JDS");
                    pt = routingDao.save(pt);
//                        String altPid = UidUtils.getLocalPatientIdFromDomainUid(item.path("uid").asText());
//                        pt = patientDao.findByAnyPid(PidUtils.getPid(null, altPid, vistaId));
//                        if (pt == null) {
//                            throw new DataRetrievalFailureException("Can't find patient by pid: " + pid + " or alternate dfn: " + altPid + " when checking async stream");
//                        }
                }
                if (pt == null) {
                    LOG.error("Cannot find patient by pid: " + pid + " Item data will be skipped for uid: " + item.path("uid").asText(), new DataRetrievalFailureException("Can't find patient by pid: " + pid + " when checking async stream"));
                    return;
                }
                patMap.put(pt.getPid(), pt);
            }
            SyncStatus stat = statMap.get(pt.getPid());
            if (stat == null) {
                stat = syncStatusDao.findOneByPid(pt.getPid());
            }
            if (stat == null) {
                stat = createSyncStatus(vistaId, pt);
            }

            // This should probably be inside the SyncStatus object.
            stat.updateDomainTotal(vistaId, domainName, tot);
            stat.updateDomainCount(vistaId, domainName, idx);

            statMap.put(pt.getPid(), stat);
            VistaDataChunk chunk = VistaDataChunk.createVistaDataChunk(vistaId, url, object, domainName, idx, tot, pt, VistaDataChunk.getProcessorParams(vistaId, pt.getLocalPatientIdForSystem(vistaId), false), isBatch);
            chunks.add(chunk);
        }
    }

    private void handleOperationalDataItem(String vistaId, List<VistaDataChunk> chunks, String url, String domainName, int tot, int idx, JsonNode object, Map<String, SyncStatus> statMap, boolean isBatch) {
        SyncStatus syncStatus = statMap.get("OPD");
        if (syncStatus == null) {
            syncStatus = syncStatusDao.findOneForOperational();
            if (syncStatus == null) {
                syncStatus = new SyncStatus();
                syncStatus.setForOperational(true);
            }
            statMap.put("OPD", syncStatus);
        }
        if (domainName.equals("syncStatus") || domainName.equals("syncStart") || domainName.equals("syncDone")) {
            if (domainName.equals("syncStatus") || domainName.equals("syncDone")) {
                syncStatus.setSyncOperationalComplete(true);
                VistaDataChunk chunk = VistaDataChunk.createVistaDataChunk(vistaId, url, POMUtils.parseJSONtoNode(syncStatus.toJSON()), domainName, idx, tot, null, VistaDataChunk.getProcessorParams(vistaId, null, false), isBatch);
                chunks.add(chunk);
            }
        } else if (tot > 0 && idx > 0) {
            if (object.path("uid").isMissingNode()) {
                LOG.warn("Skipping '" + domainName + "' operational item with missing UID.");
            } else {
                syncStatus.updateOperationalDomainTotal(domainName, tot);
                syncStatus.updateOperationalDomainCount(domainName, idx);
                VistaDataChunk chunk = VistaDataChunk.createVistaDataChunk(vistaId, url, object, domainName, idx, tot, null, VistaDataChunk.getProcessorParams(vistaId, null, false), isBatch);
                chunks.add(chunk);
            }
        }
    }

    @Override
    public void setEnvironment(Environment environment) {
        this.environment = environment;
        this.hdrEnabled = environment.getProperty(HmpProperties.HDR_ENABLED, Boolean.class, false);
        this.asyncBatchSize = environment.getProperty(HmpProperties.SYNC_BATCH_SIZE, Integer.class, DEFAULT_BATCH_SIZE);
        this.asyncMaxSizeBytes = environment.getProperty(HmpProperties.SYNC_MAX_SIZE_BYTES, Integer.class, DEFAULT_MAX_SIZE_BYTES);
        this.asyncTargetTimeSeconds = environment.getProperty(HmpProperties.SYNC_TARGET_TIME_SECONDS, Integer.class, DEFAULT_TARGET_TIME_SECONDS);
    }
}