package gov.va.cpe.vpr.sync.vista;

import org.springframework.scheduling.Trigger;
import org.springframework.scheduling.TriggerContext;
import org.springframework.util.Assert;

import java.util.Date;
import java.util.concurrent.TimeUnit;

/**
 * Trigger implementation that adjusts next execution time based on how long last execution took, attempting to run as
 * close to the period as possible. If execution took longer than period, then nextExecution time will be immediately.
 */
public class VprUpdateTrigger implements Trigger {

    private long initialDelay = 0;

    private long period;

    private TimeUnit timeUnit;

    /**
     * Create a trigger with the given period in milliseconds.
     */
    public VprUpdateTrigger(long period) {
        this(period, null);
    }

    /**
     * Create a trigger with the given period and time unit. The time unit will
     * apply not only to the period but also to any 'initialDelay' value, if
     * configured on this Trigger later via {@link #setInitialDelay(long)}.
     */
    public VprUpdateTrigger(long period, TimeUnit timeUnit) {
        Assert.isTrue(period >= 0, "period must not be negative");
        this.timeUnit = (timeUnit != null ? timeUnit : TimeUnit.MILLISECONDS);
        this.period = this.timeUnit.toMillis(period);
    }

    /**
     * Specify the delay for the initial execution. It will be evaluated in
     * terms of this trigger's {@link TimeUnit}. If no time unit was explicitly
     * provided upon instantiation, the default is milliseconds.
     */
    public void setInitialDelay(long initialDelay) {
        this.initialDelay = this.timeUnit.toMillis(initialDelay);
    }

    public long getInitialDelay() {
        return initialDelay;
    }

    public long getPeriod() {
        return period;
    }

    public TimeUnit getTimeUnit() {
        return timeUnit;
    }

    @Override
    public Date nextExecutionTime(TriggerContext triggerContext) {
        if (triggerContext.lastScheduledExecutionTime() == null) {
            return new Date(System.currentTimeMillis() + this.initialDelay);
        }
        long elapsed = triggerContext.lastCompletionTime().getTime() - triggerContext.lastActualExecutionTime().getTime();
        if (elapsed < this.period) {
            return new Date(triggerContext.lastCompletionTime().getTime() + this.period - elapsed);
        } else {
            return new Date(triggerContext.lastCompletionTime().getTime() + 1);
        }
    }
}
