package gov.va.hmp.spring;

import gov.va.cpe.vpr.frameeng.FrameEng;
import gov.va.cpe.vpr.frameeng.FrameRegistry;
import gov.va.cpe.vpr.frameeng.FrameRunner;
import gov.va.cpe.vpr.pom.IGenericPatientObjectDAO;
import gov.va.cpe.vpr.pom.IPatientDAO;
import gov.va.cpe.vpr.pom.jds.JdsOperations;
import gov.va.cpe.vpr.termeng.H2TermDataSource;
import gov.va.cpe.vpr.termeng.ITermDataSource;
import gov.va.cpe.vpr.termeng.LuceneSearchDataSource;
import gov.va.cpe.vpr.termeng.TermEng;
import gov.va.hmp.HmpProperties;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Lazy;
import org.springframework.core.env.Environment;

import java.io.IOException;
import java.sql.SQLException;

@Configuration
public class FrameConfig {

    public static final String SNOMED_CT_DB = "sctdb";
    public static final String LOINC_DB = "lncdb";
    public static final String DRUG_DB = "drugdb";
    public static final String ICD_DB = "icddb";

    @Autowired
    Environment env;

    @Autowired
    JdsOperations jdsTemplate;

    @Autowired
    IGenericPatientObjectDAO genericDao;

    @Autowired
    IPatientDAO patientDao;

    @Lazy
    @Bean
    public FrameRunner frameRunner() {
        return new FrameRunner(frameRegistry());
    }

    // Frame engine that runs async (listens to events) and saves generated alert
    @Bean
    public FrameEng frameEng() {
        return new FrameEng(frameRegistry(), jdsTemplate, genericDao, patientDao);
    }

    @Lazy
    @Bean(initMethod = "load")
    public FrameRegistry frameRegistry() {
        return new FrameRegistry(
                springFrameLoader(),
                viewDefDefLoader()
                // Not ready yet:
//                new FrameRegistry.DroolsFrameLoader(),
//                new FrameRegistry.RemindersFrameLoader(),
//                new FrameRegistry.ProtocolFileFrameLoader()
        );
    }

    @Lazy
    @Bean
    public FrameRegistry.IFrameLoader springFrameLoader() {
        return new FrameRegistry.SpringFrameLoader();
    }

    @Bean(name = "vddLoader")
    public FrameRegistry.IFrameLoader viewDefDefLoader() {
        return new gov.va.cpe.vpr.frameeng.FrameRegistry.ViewDefDefFrameLoader();
    }

    @Bean
    public TermEng termEng() {
        try {
            TermEng termEng = TermEng.createInstance(new ITermDataSource[]{
                    new LuceneSearchDataSource(termEngSCTDB(), getLuceneDir(SNOMED_CT_DB), false),
                    new LuceneSearchDataSource(termEngLNCDB(), getLuceneDir(LOINC_DB), false),
                    new LuceneSearchDataSource(termEngDrugDB(), getLuceneDir(DRUG_DB), false),
                    new LuceneSearchDataSource(termEngICDDB(), getLuceneDir(ICD_DB), false)
            });
            return termEng;
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    @Bean
    public H2TermDataSource termEngSCTDB() {
        return h2TermDataSource(SNOMED_CT_DB);
    }

    @Bean
    public H2TermDataSource termEngLNCDB() {
        return h2TermDataSource(LOINC_DB);
    }

    @Bean
    public H2TermDataSource termEngDrugDB() {
        return h2TermDataSource(DRUG_DB);
    }

    @Bean
    public H2TermDataSource termEngICDDB() {
        return h2TermDataSource(ICD_DB);
    }

    private H2TermDataSource h2TermDataSource(String db) {
        try {
            return new H2TermDataSource(getJdbcUrl(db));
        } catch (SQLException e) {
            throw new RuntimeException(e);
        } catch (ClassNotFoundException e) {
            throw new RuntimeException(e);
        }
    }

    private String getJdbcUrl(String db) {
        return "jdbc:h2:" + getHmpDataDir() + "/termdb-" + getDBVersion(db) + "-" + db + "/termdb;ACCESS_MODE_DATA=r;DB_CLOSE_ON_EXIT=FALSE";
    }

    private String getLuceneDir(String db) {
        return getHmpDataDir() + "/termdb-" + getDBVersion(db) + "-" + db + "/lucene";
    }

    private String getDBVersion(String db) {
        return env.getRequiredProperty("termdb." + db + ".version");
    }

    private String getHmpDataDir() {
        return env.getRequiredProperty(HmpProperties.HMP_DATA_DIR);
    }

    @Bean
    public FrameRunner.JDSSaveActionRunner jdsSaveActionRunner() {
        return new FrameRunner.JDSSaveActionRunner(jdsTemplate);
    }
}
