package gov.va.cpe.vpr;

import gov.va.cpe.vpr.pom.POMUtils;
import org.junit.Test;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import static org.hamcrest.CoreMatchers.is;
import static org.hamcrest.CoreMatchers.nullValue;
import static org.junit.Assert.assertThat;

public class DocumentTests {

    private Document d = new Document();

    @Test
    public void testConstruct() throws Exception {
        assertThat(d.getAuthor(), nullValue());
        assertThat(d.getCosigner(), nullValue());
        assertThat(d.getAttending(), nullValue());
    }

    @Test
    public void testGetAuthor() throws Exception {
        d = createDocumentWithClinician("FOO", DocumentClinician.Role.AUTHOR_DICTATOR);
        assertThat(d.getAuthor(), is("FOO"));
        assertThat(d.getAuthorDisplayName(), is("Foo"));
    }

    @Test
    public void testGetCosigner() throws Exception {
        d = createDocumentWithClinician("BAR", DocumentClinician.Role.COSIGNER);
        assertThat(d.getCosigner(), is("BAR"));
        assertThat(d.getCosignerDisplayName(), is("Bar"));
    }

    @Test
    public void testGetAttending() throws Exception {
        d = createDocumentWithClinician("BAZ", DocumentClinician.Role.ATTENDING_PHYSICIAN);
        assertThat(d.getAttending(), is("BAZ"));
        assertThat(d.getAttendingDisplayName(), is("Baz"));
    }

    @Test
    public void testIsTIUDocument() throws Exception {
        assertThat(Document.isTIU("urn:va:document:ABCD:229:456"), is(true));
        d.setData("uid", "urn:va:document:ABCD:229:456");
        assertThat(Document.isTIU(d), is(true));
        assertThat(Document.isTIU(POMUtils.convertObjectToMap(d)), is(true));

        assertThat(Document.isTIU("urn:va:document:ABCD:229:456;789"), is(false));
        d.setData("uid", "urn:va:document:ABCD:229:456;789");
        assertThat(Document.isTIU(d), is(false));
        assertThat(Document.isTIU(POMUtils.convertObjectToMap(d)), is(false));

        assertThat(Document.isTIU("urn:va:foo:ABCD:229:456"), is(false));
        d.setData("uid", "urn:va:foo:ABCD:229:456");
        assertThat(Document.isTIU(d), is(false));
        assertThat(Document.isTIU(POMUtils.convertObjectToMap(d)), is(false));
    }

    @Test
    public void testGetStatusNameWithDocumentInstance() throws Exception {
        Document document = new Document();
        document.setData("status", "completed");

        assertThat(Document.getStatusName(document), is("completed"));
        assertThat(Document.getStatusName(POMUtils.convertObjectToMap(document)), is("completed"));
    }

    @Test
    public void testGetStatusNameWithSolrSearchResult() throws Exception {
        Map<String, Object> document = new HashMap<>();
        document.put(Document.SOLR_DOCUMENT_STATUS_FIELD, Arrays.asList("retracted"));

        assertThat(Document.getStatusName(document), is("retracted"));
    }

    @Test
    public void testGetDocDefUidWithDocumentInstance() throws Exception {
        Document document = new Document();
        document.setData("documentDefUid", "urn:va:doc-def:ABCD:34");

        assertThat(Document.getDocDefUid(document), is("urn:va:doc-def:ABCD:34"));
        assertThat(Document.getDocDefUid(POMUtils.convertObjectToMap(document)), is("urn:va:doc-def:ABCD:34"));
    }

    @Test
    public void testGetDocDefUidWithSolrSearchResult() throws Exception {
        Map<String, Object> document = new HashMap<>();
        document.put(Document.SOLR_DOC_DEF_UID_FIELD, Arrays.asList("urn:va:doc-def:ABCD:34"));

        assertThat(Document.getDocDefUid(document), is("urn:va:doc-def:ABCD:34"));
    }

    @Test
    public void testGetLocalTitleWithDocumentInstance() throws Exception {
        Document document = new Document();
        document.setData("localTitle", "Foo Bar Baz");

        assertThat(Document.getLocalTitle(document), is("Foo Bar Baz"));
        assertThat(Document.getLocalTitle(POMUtils.convertObjectToMap(document)), is("Foo Bar Baz"));
    }

    @Test
    public void testGetLocalTitleWithSolrSearchResult() throws Exception {
        Map<String, Object> document = new HashMap<>();
        document.put(Document.SOLR_LOCAL_TITLE_FIELD, Arrays.asList("Foo Bar Baz"));

        assertThat(Document.getLocalTitle(document), is("Foo Bar Baz"));
    }

    @Test
    public void testGetAuthorUidWithDocumentInstance() throws Exception {
        Document document = new Document();
        document.setData("authorUid", UidUtils.getUserUid("ABCD", "23"));

        assertThat(Document.getAuthorUid(document), is(UidUtils.getUserUid("ABCD", "23")));
        assertThat(Document.getAuthorUid(POMUtils.convertObjectToMap(document)), is(UidUtils.getUserUid("ABCD", "23")));
    }

    @Test
    public void testGetAuthorUidWithSolrSearchResult() throws Exception {
        Map<String, Object> document = new HashMap<>();
        document.put(Document.SOLR_AUTHOR_UID_FIELD, Arrays.asList(UidUtils.getUserUid("ABCD", "23")));

        assertThat(Document.getAuthorUid(document), is(UidUtils.getUserUid("ABCD", "23")));
    }

    @Test
    public void testGetSignerUidWithDocumentInstance() throws Exception {
        Document document = new Document();
        document.setData("signerUid", UidUtils.getUserUid("ABCD", "23"));

        assertThat(Document.getSignerUid(document), is(UidUtils.getUserUid("ABCD", "23")));
        assertThat(Document.getSignerUid(POMUtils.convertObjectToMap(document)), is(UidUtils.getUserUid("ABCD", "23")));
    }

    @Test
    public void testGetSignerUidWithSolrSearchResult() throws Exception {
        Map<String, Object> document = new HashMap<>();
        document.put(Document.SOLR_SIGNER_UID_FIELD, Arrays.asList(UidUtils.getUserUid("ABCD", "23")));

        assertThat(Document.getSignerUid(document), is(UidUtils.getUserUid("ABCD", "23")));
    }

    @Test
    public void testGetCosignerUidWithDocumentInstance() throws Exception {
        Document document = new Document();
        document.setData("cosignerUid", UidUtils.getUserUid("ABCD", "23"));

        assertThat(Document.getCosignerUid(document), is(UidUtils.getUserUid("ABCD", "23")));
        assertThat(Document.getCosignerUid(POMUtils.convertObjectToMap(document)), is(UidUtils.getUserUid("ABCD", "23")));
    }

    @Test
    public void testGetCosignerUidWithSolrSearchResult() throws Exception {
        Map<String, Object> document = new HashMap<>();
        document.put(Document.SOLR_COSIGNER_UID_FIELD, Arrays.asList(UidUtils.getUserUid("ABCD", "23")));

        assertThat(Document.getCosignerUid(document), is(UidUtils.getUserUid("ABCD", "23")));
    }

    @Test
    public void testGetAttendingUidWithDocumentInstance() throws Exception {
        Document document = new Document();
        document.setData("attendingUid", UidUtils.getUserUid("ABCD", "23"));

        assertThat(Document.getAttendingUid(document), is(UidUtils.getUserUid("ABCD", "23")));
        assertThat(Document.getAttendingUid(POMUtils.convertObjectToMap(document)), is(UidUtils.getUserUid("ABCD", "23")));
    }

    @Test
    public void testGetAttendingUidWithSolrSearchResult() throws Exception {
        Map<String, Object> document = new HashMap<>();
        document.put(Document.SOLR_ATTENDING_UID_FIELD, Arrays.asList(UidUtils.getUserUid("ABCD", "23")));

        assertThat(Document.getAttendingUid(document), is(UidUtils.getUserUid("ABCD", "23")));
    }

    public static Document createDocumentWithClinician(String clinicianName, DocumentClinician.Role clinicianRole) {
    	String uid = "urn:va:F484:fee:;fi:fo:fum";
        DocumentText text = DocumentTextTests.createDocumentTextWithClinician(uid, clinicianName, clinicianRole);
    	Document doc = new Document();
    	doc.setData("uid",uid);
        doc.setData("text", Collections.singletonList(text));
        return doc;
    }
}

