package gov.va.cpe.vpr;

import static org.junit.Assert.*;
import gov.va.cpe.vpr.IBroadcastMessageFilter.PIDMessageFilter;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicInteger;

import org.apache.activemq.ActiveMQConnectionFactory;
import org.apache.lucene.queries.function.valuesource.DefFunction;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import org.springframework.web.context.request.async.DeferredResult;
import org.springframework.web.context.request.async.DeferredResult.DeferredResultHandler;
import org.springframework.web.servlet.ModelAndView;

import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.JMSException;

public class JMSBroadcastServiceTests {
	private static ConnectionFactory FACT;
	private static Connection CONN;
	private static JMSBroadcastService svc;
	
	@BeforeClass
	public static void beforeclass() throws JMSException {
		// setup JMS topic/connection/session
		String brokerURL = "vm://hmp-test-broadcast?waitForStart=1000&broker.persistent=false&broker.deleteAllMessagesOnStartup=true&broker.useJmx=true";
		FACT = new ActiveMQConnectionFactory(brokerURL);
		CONN = FACT.createConnection();
		CONN.start();
		svc = new JMSBroadcastService();
		svc.fact = FACT;
		svc.initJMSConsumer();
	}
	
	public static class TestResultHandler<T> implements DeferredResult.DeferredResultHandler {
		private T ret;
		private AtomicInteger counter;
		
		public TestResultHandler(AtomicInteger counter) {
			this.counter=counter;
		}
		
		public TestResultHandler() {
			// TODO Auto-generated constructor stub
		}
		
		@Override
		public void handleResult(Object result) {
			if (result instanceof ModelAndView) {
				ModelAndView mv = (ModelAndView) result;
				Object obj = mv.getModelMap().get("data");
				if (obj instanceof List) {
					ret = (T) ((List) obj).get(0);
				} else {
					ret = (T) obj;
				}
			}
		}
		
		public T getData() {
			return ret;
		}
	}
	
	@Test
    public void basicTest() throws JMSException, InterruptedException {
		String msg = new String("hello world");
		
		// register listener with 0 timeout, should return immediately
		TestResultHandler<String> handler = new TestResultHandler<>();
		DeferredResult<ModelAndView> result = new DeferredResult<>(250L, "[]");
		result.setResultHandler(handler);
		svc.registerUIListener("sess-1234", "client-1234", result, null);
		
		// send something
		svc.broadcastMessage(msg);
		
		// will not be immediately available due to eager fetch
		assertNull(handler.getData());
		assertFalse(result.isSetOrExpired());
		
		// wait 300ms
		Thread.currentThread().sleep(500);
		
		// result should be available now
		assertTrue(result.isSetOrExpired());
		Object data = handler.getData();
		
		// should equal the input
		assertEquals(msg, data);
		assertSame(msg, data);
    }
	
	// TODO: Test the eager fetch
	// TODO: Test a DeferredResult with 0 timeout (should not need to sleep the thread)
	// TODO: Test sending a message with a TTL
	// TODO: test at the controller level
	
	@Test
	public void testPIDTarget() throws JMSException, InterruptedException {
		// register listener with 0 timeout, should return immediately
		TestResultHandler<String> handler = new TestResultHandler<>();
		DeferredResult<ModelAndView> result = new DeferredResult<>(10L, "[]");
		result.setResultHandler(handler);
		PIDMessageFilter filter = new PIDMessageFilter("666");
		svc.registerUIListener("sess-1234", "client-1234", result, filter);
		
		// send message for patient 666
		String msg = "fi fy fo fum";
		Map<String, Object> headers = new HashMap<>();
		headers.put("pid", "666");
		svc.broadcastMessage(msg, headers);
		
		Thread.currentThread().sleep(500);
		
		// should get it
		assertEquals(msg, handler.getData());
	}
	
	@Test
	public void testPIDFilter() throws JMSException, InterruptedException {
		// register listener with 0 timeout, should return immediately
		TestResultHandler<String> handler = new TestResultHandler<>();
		DeferredResult<ModelAndView> result = new DeferredResult<>(10L, "[]");
		result.setResultHandler(handler);
		PIDMessageFilter filter = new PIDMessageFilter("666");
		svc.registerUIListener("sess-1234", "client-1234", result, filter);
		
		// send message for patient 229
		String msg = "fi fy fo fum";
		Map<String, Object> headers = new HashMap<>();
		headers.put("pid", "229");
		svc.broadcastMessage(msg, headers);
		
		Thread.currentThread().sleep(500);
		
		// should NOT get it
		assertNull(msg, handler.getData());
	}
	
	/** Make sure that we dont run out of memory or thread limits */
	@Test
	public void testForMemLeak() throws JMSException {
		for (int i = 0; i < 1000; i++) {
			svc.closeConsumers(null);

			for (int j = 0; j < 1000; j++) {
				svc.broadcastMessage("msg" + j);
			}
		}

	}
	
	/**
	 * Test sending 1M messages through the queue in small (25k) batches and ensure that it does not
	 * consume excessive memory
	 * TODO:This test currently only works when run independently, not as part of test suite. 
	 */
//	@Test
	public void testBulkMessages() throws JMSException, InterruptedException {
		final AtomicInteger consumed = new AtomicInteger(0);
		final AtomicInteger invoked = new AtomicInteger(0);
	
		// register a listener
		DeferredResult<ModelAndView> result = new DeferredResult<>(100L, "[]");
		DeferredResultHandler handler = new DeferredResultHandler() {
			@Override
			public void handleResult(Object results) {
				invoked.incrementAndGet();
				if (results instanceof ModelAndView) {
					ModelAndView view = (ModelAndView) results;
					ArrayList<String> list = (ArrayList<String>) view.getModelMap().get("data");
					consumed.addAndGet(list.size());
				}
			}
		};
		result.setResultHandler(handler);
		svc.registerUIListener("sess-0", "client-0", result, null);
//		svc.minEagerFetchTimeMS = 30_000;
//		svc.maxMessagesPerResponse = 1;
		
		// measure heap space
		System.gc();
		long heapBefore = Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory();
		
		// send lots of messages in multiple batches (to simulate many sessions/clients)
		int produced=0;
		for (; produced <= 1_000_000; produced++) {
			svc.broadcastMessage("MSG " + produced, false, 10000000, null);
			if (produced % 20000 == 0) {
				while (consumed.get() < produced) {
					// re-register the UIListener for more
					result = new DeferredResult<>(100L, "[]");
					result.setResultHandler(handler);
					svc.registerUIListener("sess-0", "client-0", result, null);
				}
			}
		}
		
		// measure heap space again
		System.gc();
		long heapAfter = Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory();
		//System.out.println("Heap Before: " + heapBefore + "; heap After: " + heapAfter);
		//System.out.println("Diff: " + (heapBefore-heapAfter));

		// consumed and produced should be same
		assertEquals(produced, consumed.get());

		// should be roughly the same (10M fudge factor)
		assertEquals(0, heapBefore-heapAfter, 10_000_000);
	}

	/** Simulate many sessions/clients while sending/reciting messages */
	@Test
	public void testSessions() {
		// TODO: Implement this
	}

}
