package gov.va.cpe.vpr;

import gov.va.cpe.vpr.Medication.MedicationKind;
import gov.va.cpe.vpr.pom.IPOMObject;
import gov.va.cpe.vpr.pom.POMObjectEvent;
import gov.va.cpe.vpr.pom.POMUtils;
import gov.va.cpe.vpr.termeng.Concept;
import gov.va.cpe.vpr.termeng.TermEng;
import gov.va.hmp.healthtime.PointInTime;
import org.junit.Before;
import org.junit.Test;

import java.io.File;
import java.io.FileNotFoundException;
import java.net.URISyntaxException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Map;

import static org.junit.Assert.*;
import static org.mockito.Matchers.anyString;
import static org.mockito.Mockito.*;

public class MedicationTests {
	
	// a simple and complex IV med to play with
	private static File med10121 = POMObjectTester.resolveFile("med10121.json", MedicationTests.class);
	private static File med34973 = POMObjectTester.resolveFile("med34973.json", MedicationTests.class);
	
	// complex inpatient med
	private static File med37504 = POMObjectTester.resolveFile("med37504.json", MedicationTests.class);
	
	private static POMObjectTester<Medication> LOADER = new POMObjectTester<>();
	
	private TermEng mockTermEng;
	private Concept mockSCLConcept;
	private HashSet<String> sclAncestorSet;
	private HashSet<String> sclEquivSet;
	private Medication med;

	@Before
	public void setup() {
		sclEquivSet = new HashSet<>();
		sclAncestorSet = new HashSet<>();
		for (int i=1; i <= 25; i++) sclAncestorSet.add("urn:ancestor:" + i);
		for (int i=1; i <= 25; i++) sclAncestorSet.add("urn:equiv:" + i);
		
		// when searching for SODIUM CHLORIDE, return a mock with NDRT mappings, 25 mocked ancestors and 25 mocked ancestors
		TermEng.setInstance(mockTermEng = mock(TermEng.class));
		mockSCLConcept = mock(Concept.class);
		when(mockSCLConcept.getMappingTo("ndfrt")).thenReturn(mockSCLConcept);
		when(mockSCLConcept.getAncestorSet()).thenReturn(sclAncestorSet);
		when(mockSCLConcept.getEquivalentSet()).thenReturn(sclEquivSet);
		when(mockTermEng.getConcept("urn:vandf:4017444")).thenReturn(mockSCLConcept);
	}
	
	@Test
	public void testRXNCodeCalculation() throws FileNotFoundException, URISyntaxException {
		// med1021 has 500+ codes and was taking a long time to return
		Map<String, Object> data = LOADER.loadFileData(med10121); 
		assertEquals("urn:va:med:F484:8:10121", data.get("uid"));
		
		// if we remove the rxncodes value from the data, some should be generated
		data.remove("rxncodes");
		med = POMUtils.newInstance(Medication.class, data);
		assertNull(med.rxnCodes);
		assertEquals(51, med.getRXNCodes().size()); // 25 ancestors, 25 equivs plus self
		verify(mockTermEng).getConcept("urn:vandf:4017444");
	}
	
	
	@Test
	public void testRXNCodeNotCalculated() throws FileNotFoundException, URISyntaxException {
		Map<String, Object> data = LOADER.loadFileData(med10121); 
		assertEquals("urn:va:med:F484:8:10121", data.get("uid"));
		
		// if we keep the rxncodes value from the data, it should be the new value and should be instant
		data.put("rxncodes", Arrays.asList("FOO","BAR","BAZ"));
		med = POMUtils.newInstance(Medication.class, data);
		assertEquals(3, med.rxnCodes.size());
		assertEquals(3, med.getRXNCodes().size());
		assertTrue(med.getRXNCodes().contains("FOO"));
		assertTrue(med.getRXNCodes().contains("BAR"));
		assertTrue(med.getRXNCodes().contains("BAZ"));
		
		// the termeng should not be called
		verify(mockTermEng, never()).getConcept(anyString());
	}
	
	@Test
	public void testRXNShouldTriggerEvents() throws FileNotFoundException, URISyntaxException {
		Medication med = LOADER.loadFile(med10121);
		
		List<POMObjectEvent<IPOMObject>> events = med.getEvents();
		assertFalse(events.isEmpty());
		
		// altering some fields produce more events
		med.setData("dose", "FOO");
		events = med.getEvents();
		assertEquals(1, events.size());
		med.clearEvents();
		
		// altering the RXN code produces events
		med.setData("rxncodes", Arrays.asList("FOO", "BAR"));
		events = med.getEvents();
		assertEquals(1, events.size());
	}

	@Test
	public void testMedEventHasPid() {
		Medication med = LOADER.loadFile(med10121);

		List<POMObjectEvent<IPOMObject>> events = med.getEvents();
		assertFalse(events.isEmpty());

		// altering some fields produce more events
		med.setData("dose", "FOO");
		events = med.getEvents();
		assertEquals(1, events.size());
		assert(events.get(0) instanceof POMObjectEvent);
	}
	
	@Test
	public void testIsCurrent() throws FileNotFoundException, URISyntaxException {
		Medication med = LOADER.loadFile(med10121);
		// treat as outpatient
		med.setData("vaType","O");
		
		// its expired and old, therefor not current
		assertEquals("EXPIRED", med.getVaStatus());
		assertTrue(med.getOverallStop().before(new PointInTime("2000")));
		assertTrue(med.getStopped().before(new PointInTime("2000")));
		assertFalse(med.isRecent());
		
		// setting to active, pending, hold, suspend should make it current
		med.setData("vaStatus", "ACTIVE");
		assertTrue(med.isRecent());
		med.setData("vaStatus", "PENDING");
		assertTrue(med.isRecent());
		med.setData("vaStatus", "HOLD");
		assertTrue(med.isRecent());
		med.setData("vaStatus", "SUSPEND");
		assertTrue(med.isRecent());
		
		// setting status to EXPIRED, DISCONTINED should not be current
		med.setData("vaStatus", "EXPIRED");
		assertFalse(med.isRecent());
		med.setData("vaStatus", "DISCONTINUED");
		assertFalse(med.isRecent());
		med.setData("vaStatus", "DISCONTINUED/EDIT");
		assertFalse(med.isRecent());
		
		// setting stopped date to within 30 days should make it current
		med.setData("vaStatus", "EXPIRED");
		med.setData("overallStop", PointInTime.today().subtractDays(30));
		med.setData("stopped", PointInTime.today().subtractDays(30));
		assertTrue(med.isRecent());
	}
	
	@Test
	public void testIsBetween() throws FileNotFoundException, URISyntaxException {
		Medication med = LOADER.loadFile(med10121);
		
		// nulls always return false
		assertFalse(med.isBetween(null, PointInTime.now()));
		assertFalse(med.isBetween(PointInTime.now(), null));
		
		// this is the start and stop date
		assertEquals("199906281500", med.getOverallStart().toString());
		assertEquals("199907012359", med.getOverallStop().toString());
		
		// return true if the start date falls inbetween the range, or stop date falls between range
		assertTrue(med.isBetween(new PointInTime("19990625"), new PointInTime("19900000")));
		assertTrue(med.isBetween(new PointInTime("19900000"), new PointInTime("19990705")));
		
		// same day and missmatched accuracy should return true as well
		assertTrue(med.isBetween(new PointInTime("19990628"), new PointInTime("19990628")));
		assertTrue(med.isBetween(new PointInTime("19990701"), new PointInTime("19990701")));
	}
	
	@Test
	public void testDailyDose() {
		Medication med = LOADER.loadFile(med37504);
		assertEquals(1, med.getDosages().size());
		MedicationDose dose = med.getDosages().get(0);
				
		// 2mg Q4H PRN = 12mg/daily, should return as integer
		assertEquals(12, med.getDailyDose());
		assertTrue(dose.calcDailyDose().floatValue() == med.getDailyDose().floatValue());
		
		// .5mg Q4H = 3mg/daily, should return as float
		dose.setData("dose", .5d);
		assertEquals(3f, dose.calcDailyDose().floatValue(), 0);
		assertTrue(dose.calcDailyDose().floatValue() == med.getDailyDose().floatValue());
		
		// 1.5mg Q8H = 4.5mg/daily, should return as float
		dose.setData("dose", 1.5d);
		dose.setData("scheduleFreq", 60*8);
		assertEquals(4.5f, dose.calcDailyDose().floatValue(), 0);
		assertTrue(dose.calcDailyDose().floatValue() == med.getDailyDose().floatValue());
		
		// for Medication/Dose, test exactly once daily returns exact dose (DrugTherapy is different)
		dose.setData("scheduleFreq", 1440);
		assertEquals(1.5f, dose.calcDailyDose().floatValue(), 0);
		assertTrue(dose.calcDailyDose().floatValue() == med.getDailyDose().floatValue());
		
		// test less than once every other day returns .75 (DrugTherapy is different)
		dose.setData("scheduleFreq", 2880);
		assertEquals(0.75f, dose.calcDailyDose().floatValue(), 0);
		assertTrue(dose.calcDailyDose().floatValue() == med.getDailyDose().floatValue());
		
	}
	
	@Test
	public void testDoseAdmin() {
		
		Medication med = LOADER.loadFile(med10121);
		assertEquals(1, med.getDosages().size());
		
		// check the dose fields
		MedicationDose dose = med.getDosages().get(0);
		assertEquals("20 ml/hr", dose.getIvRate());
		assertEquals("IV", dose.getRouteName());
		assertEquals("BID", dose.getScheduleName());
		assertEquals("CONTINUOUS", dose.getScheduleType());
		assertEquals(new Integer(720), dose.getScheduleFreq());
		assertEquals("0700-1900", dose.getAdminTimes());
		
		// check dose administrations, should be returned in a sorted order
		assertEquals(3, med.getAdministrations().size());
		
		// last admin only includes GIVEN + INFUSING, so getLastAdmin() is not necessarily == getAdminitrations.first
		assertEquals("NOT GIVEN", med.getAdministrations().first().getStatus());
		assertFalse(med.getLastAdmin().equals(med.getAdministrations().first().getDateTime()));
		
		MedicationAdministration admin = med.getAdministrations().first();
		assertEquals("200006281900", admin.getDateTime().toString());
		assertEquals("NURSE,TWENTY-SIX", admin.getAdministeredByName());
		assertEquals("urn:va:user:F484:20189", admin.getAdministeredByUid());
		assertEquals("NOT GIVEN", admin.getStatus());
		
		admin = med.getAdministrations().last();
		assertEquals("199906281900", admin.getDateTime().toString());
		assertEquals("NURSE,TWENTY-SIX", admin.getAdministeredByName());
		assertEquals("urn:va:user:F484:20189", admin.getAdministeredByUid());
		assertEquals("GIVEN", admin.getStatus());
		
		// each administration should have 1+ medications that were administered
		for (MedicationAdministration medAdmin : med.getAdministrations()) {
			assertNotNull(medAdmin.getMedications());
			assertEquals(1, medAdmin.getMedications().size());
			MedicationAdministrationMed adminMed = medAdmin.getMedications().get(0);
			assertEquals("SODIUM CHLORIDE 0.9% INJ", adminMed.getName());
			assertEquals("INJ", adminMed.getUnits());
			assertEquals("1", adminMed.getAmount());
		}
		
		// next dose should either be at 0700 or 1900
		// at high noon, the next dose admin should be 1900
		PointInTime now = new PointInTime("199906291200");
		assertEquals(new PointInTime("199906291900"), med.calcNextScheduledAdminFrom(now));
		
		// if the admin times were single numbers, ensure that it still calculates the full hour/minute
		med.getDosages().get(0).setData("adminTimes", "04-08-12-16-20"); // initate Q4H
		now = new PointInTime("199906290500"); // next dose should be at 0800 hours
		assertEquals(new PointInTime("199906290800"), med.calcNextScheduledAdminFrom(now));
	}
	
	@Test
	public void testComplexIV() {
		Medication med = LOADER.loadFile(med34973);
		assertEquals(1, med.getDosages().size());
		assertEquals(2, med.getProducts().size());
		assertEquals(12, med.getAdministrations().size());
		
		// cannot currently calculate dose admin for IV's
		PointInTime start = new PointInTime("201403060000");
		PointInTime end = new PointInTime("2014030PORT0");
		assertNull(med.calcDoseAdminBetween(start, end));
		
		// IV doesn't currently compute daily dose
		assertNull(med.getDailyDose());
	}
	
	@Test
	public void testComplexINPT() {
		Medication med = LOADER.loadFile(med37504);
		assertEquals(1, med.getDosages().size());
		assertEquals(1, med.getProducts().size());
		assertEquals(3, med.getAdministrations().size());
		assertTrue(med.isPRN());
		
		// calc 24h dose over whole period is 6 mg
		PointInTime start = new PointInTime("201403060000");
		PointInTime end = new PointInTime("2014030PORT0");
		assertEquals(6.0f, med.calcDoseAdminBetween(start, end));
		
		// there were 2 admins after noon on 20140306 for a total of 4mg
		start = new PointInTime("201403061200");
		end = new PointInTime("201403071200");
		assertEquals(4.0f, med.calcDoseAdminBetween(start, end));
		
		// daily dose for 2mg Q4H is 12 
		assertEquals(new Integer(12), med.getDailyDose());
		
		// since its PRN there is no scheduled time
		assertNull(med.calcNextScheduledAdminFrom(end));
	}
	
	@Test
	public void testMedicationKind() {
		Medication med = new Medication();
		assertSame(MedicationKind.UNKNOWN, MedicationKind.kindOf(med));
		
		med.setData("vaType", "I");
		assertSame(MedicationKind.I, MedicationKind.kindOf(med));

		med.setData("vaType", "O");
		assertSame(MedicationKind.O, MedicationKind.kindOf(med));

		med.setData("vaType", "N");
		assertSame(MedicationKind.N, MedicationKind.kindOf(med));

		med.setData("vaType", "V");
		assertSame(MedicationKind.V, MedicationKind.kindOf(med));

		med.setData("supply", true);
		assertSame(MedicationKind.SUPPLY, MedicationKind.kindOf(med));
		med.setData("supply", null);
		
		med.setData("IMO", true);
		assertSame(MedicationKind.IMO, MedicationKind.kindOf(med));
	}
}
