package gov.va.cpe.vpr;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import gov.va.cpe.vpr.JMSBroadcastServiceTests.TestResultHandler;

import java.lang.management.ManagementFactory;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.jms.JMSException;
import javax.management.JMException;
import javax.management.ObjectName;

import org.junit.Test;
import org.springframework.web.context.request.async.DeferredResult;
import org.springframework.web.context.request.async.DeferredResult.DeferredResultHandler;
import org.springframework.web.servlet.ModelAndView;

public class MemBroadcastServiceTests {
	
	@Test
	public void simpleTest() {
		MemBroadcastService svc = new MemBroadcastService();
		
		String msg = "HELLO WORLD";
		String jsessid = "FOO";
		String clientid = "BAR";
		assertEquals(0, svc.getQueueCount());
		
		// does nothing, goes nowhere
		svc.broadcastMessage(msg);
		assertEquals(0, svc.getQueueCount());
		assertEquals(0, svc.getConsumeCount());
		assertEquals(0, svc.getEnqueueCount());
		assertEquals(1, svc.getBroadcastCount());

		// register a deferred result
		TestResultHandler<String> handler = new TestResultHandler<>();
		DeferredResult<ModelAndView> result = new DeferredResult<>();
		result.setResultHandler(handler);
		svc.registerUIListener(jsessid, clientid, result, null);
		
		// should be a queue for it
		assertEquals(1, svc.getQueueCount());
		assertFalse(result.isSetOrExpired());
		
		// send a message, handler should get it
		svc.broadcastMessage(msg);
		assertEquals(0, svc.getExpiredCount());
		assertEquals(1, svc.getEnqueueCount());
		assertEquals(2, svc.getBroadcastCount());
		assertEquals(1, svc.getConsumeCount());
		assertTrue(result.isSetOrExpired());
		assertEquals(msg, handler.getData());
	}
	
	@Test
	public void testFilter() {
		MemBroadcastService svc = new MemBroadcastService();
		String jsessid = "FOO";
		
		// handler 1 gets everything
		TestResultHandler<String> handler1 = new TestResultHandler<>();
		DeferredResult<ModelAndView> result1 = new DeferredResult<>();
		result1.setResultHandler(handler1);
		svc.registerUIListener(jsessid, "A", result1, null);
		
		// handler 2 gets only PID=555 messages
		TestResultHandler<String> handler2 = new TestResultHandler<>();
		DeferredResult<ModelAndView> result2 = new DeferredResult<>();
		result2.setResultHandler(handler2);
		IBroadcastMessageFilter filter2 = new IBroadcastMessageFilter.PIDMessageFilter("555");
		svc.registerUIListener(jsessid, "B", result2, filter2);

		// send message for PID=555
		Map<String, Object> headers = new HashMap<>();
		headers.put("pid", "555");
		assertEquals(0, svc.getBroadcastCount());
		assertEquals(0, svc.getConsumeCount());
		assertEquals(2, svc.getQueueCount());
		svc.broadcastMessage("MSG TO 555", headers);
		
		// both consumers get it
		assertEquals(1, svc.getBroadcastCount());
		assertEquals(2, svc.getEnqueueCount());
		assertEquals(2, svc.getConsumeCount());
		assertTrue(result1.isSetOrExpired());
		assertTrue(result2.isSetOrExpired());
		assertEquals("MSG TO 555", handler1.getData());
		assertEquals("MSG TO 555", handler2.getData());
		
		// send another message wo/PID, should get queued in 1 queue, not the other
		svc.broadcastMessage("MSG NO PID");
		assertEquals(2, svc.getBroadcastCount());
		assertEquals(3, svc.getEnqueueCount());
		assertEquals(1, svc.getQueueSize(jsessid + ":A"));
		assertEquals(0, svc.getQueueSize(jsessid + ":B"));
		
		// re-register listener1, should immediately get its result
		handler1 = new TestResultHandler<>();
		result1 = new DeferredResult<>();
		result1.setResultHandler(handler1);
		svc.registerUIListener(jsessid, "A", result1, null);
		assertEquals(0, svc.getQueueSize(jsessid + ":A"));
		assertTrue(result1.isSetOrExpired());
		assertEquals("MSG NO PID", handler1.getData());

		handler2 = new TestResultHandler<>();
		result2 = new DeferredResult<>();
		result2.setResultHandler(handler2);
		filter2 = new IBroadcastMessageFilter.PIDMessageFilter("555");
		svc.registerUIListener(jsessid, "B", result2, filter2);
		assertEquals(0, svc.getQueueSize(jsessid + ":B"));
		assertFalse(result2.isSetOrExpired());
	}
	
	public static boolean DO_DEBUG=false;
	@Test
	public void testEagerFetch() throws InterruptedException {
		MemBroadcastService svc = new MemBroadcastService();
		Thread t = null;
		Consumer c = new Consumer(svc, "FOO", "BAR", 500, null);
		try {
			t = c.start();
			Thread.sleep(1000);
			assertEquals(1, svc.getActiveDeferredResultCount());
			assertEquals(1, svc.getQueueCount());
			assertEquals(0, svc.getQueueSize("FOO:BAR"));
			
			// should be consumed right away
			svc.broadcastMessage("MSG0");
			assertEquals(0, svc.getActiveDeferredResultCount());
			assertEquals(1, c.getCount());
			assertEquals(1, c.getItems().size());
			assertEquals("MSG0", c.getItems().get(0));
			assertEquals(1, svc.getQueueCount());
			assertEquals(0, svc.getQueueSize("FOO:BAR"));
			
			// send 5 messages
			for (int i=1; i<=5; i++) svc.broadcastMessage("MSG"+i);
			DO_DEBUG=true;
			
			// should all be picked up at once in next cycle
			Thread.sleep(510);
			assertEquals(6, c.getCount());
			assertEquals("MSG5", c.getItems().get(5));
		} finally {
			c.shutdown();
			t.join();
		}
	}
	
	@Test
	public void testMultiThreaded() {
		final int PRODUCE_LIMIT = 1000;
		

//		Producer producer = new Producer(svc, PRODUCE_LIMIT, 1);
		
		
	}
	
	@Test
	public void testFullQueue() {
		
		TestResultHandler<String> handler = new TestResultHandler<>();
		DeferredResult<ModelAndView> result = new DeferredResult<>();
		result.setResultHandler(handler);
//		svc.registerUIListener(jsessid, "A", result, null);
		
		
	}

	@Test
	public void testExpired() throws InterruptedException {
		MemBroadcastService svc = new MemBroadcastService();
		
		// add listener
		TestResultHandler<String> handler = new TestResultHandler<>();
		DeferredResult<ModelAndView> result = new DeferredResult<>();
		result.setResultHandler(handler);
		svc.registerUIListener("FOO", "BAR", result, null);
		
		// first message goes right through
		assertFalse(result.isSetOrExpired());
		svc.broadcastMessage("MSG0");
		assertTrue(result.isSetOrExpired());
		assertEquals(0, svc.getExpiredCount());
		
		// send 10 more, with 10ms TTL
		for (int i=1; i <= 10; i++) svc.broadcastMessage("MSG"+i, false, 10, null);
		
		// wait 20ms, 10 messages are technically expired, but not marked as such yet
		Thread.sleep(20);
		assertEquals(0, svc.getExpiredCount());
		
		// re-attach listener
		handler = new TestResultHandler<>();
		result = new DeferredResult<>();
		result.setResultHandler(handler);
		svc.registerUIListener("FOO", "BAR", result, null);
		
		// now all 10 should be expired, nothing should be resulted
		assertEquals(10, svc.getExpiredCount());
		assertFalse(result.isSetOrExpired());
		
		
	}

	private class Producer implements Runnable {
		private int limit = 0;
		private int count = 0;
		private IBroadcastService svc;
		private int delayMS;
		
		public Producer(IBroadcastService svc, int limit, int delayMS) {
			this.limit = limit;
			this.delayMS = delayMS;
			this.svc = svc;
		}
		
		public void run() {
			while (++count < limit) {
				svc.broadcastMessage("MSG: " + count);
				try {
					if (this.delayMS > 0) { Thread.sleep(this.delayMS); }
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		}
	}
	
	private class Consumer implements DeferredResultHandler, Runnable {
		
		private List<Object> items = new ArrayList<>();
		private IBroadcastService svc;
		private String clientid;
		private String jsessid;
		private IBroadcastMessageFilter filter;
		private int reconnectDelayMS = 100;
		private boolean shutdown;
		private Thread thread;
		private int timeoutMS;

		public Consumer(IBroadcastService svc, String jsessid, String clientid, int timeoutMS, IBroadcastMessageFilter filter) {
			this.svc = svc;
			this.timeoutMS = timeoutMS;
			this.jsessid = jsessid;
			this.clientid = clientid;
			this.filter = filter;
		}
		
		public Thread start() throws InterruptedException {
			Thread t = new Thread(this);
			t.start();
			return t;
		}

		@Override
		public void run() {
			this.shutdown = false;
			this.thread = Thread.currentThread();
			
			while (!shutdown) {
				// register a deferred result
				try {
					DeferredResult<ModelAndView> result = new DeferredResult<>(this.timeoutMS);
					result.setResultHandler(this);
					svc.registerUIListener(jsessid, clientid, result, filter);
					
				} catch (JMSException e) {
					e.printStackTrace();
				}

				try {
					if (!shutdown)	Thread.sleep(this.timeoutMS);
				} catch (InterruptedException e) {
					// expected
				}
			}
			
		}
		
		public void shutdown() {
			this.shutdown = true;
			this.thread.interrupt();
		}

		@Override
		public void handleResult(Object result) {
			ModelAndView mv = (ModelAndView) result;
			Object data = mv.getModel().get("data");
			if (data != null && data instanceof Collection) {
				items.addAll((Collection) data);
			}
			
//			// interrupt thread
//			try {
//				Thread.sleep(this.reconnectDelayMS);
//			} catch (InterruptedException ex) {
//				 // ignore
//			}
//			thread.interrupt();
		}
		
		public List<Object> getItems() {
			return items;
		}
		
		public int getCount() {
			return items.size();
		}
		
	}

}
