package gov.va.cpe.vpr.sync.vista;

import org.junit.Before;
import org.junit.Test;
import org.springframework.scheduling.TriggerContext;

import java.util.Date;
import java.util.concurrent.TimeUnit;

import static org.hamcrest.CoreMatchers.*;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

public class VprUpdateTriggerTests {

    private VprUpdateTrigger t;
    private TriggerContext mockTriggerContext;

    @Before
    public void setUp() throws Exception {
        t = new VprUpdateTrigger(2, TimeUnit.SECONDS);
        t.setInitialDelay(10);
        mockTriggerContext = mock(TriggerContext.class);
    }

    @Test
    public void testConstruct() throws Exception {
        assertThat(t.getPeriod(), is(TimeUnit.SECONDS.toMillis(2)));

        t = new VprUpdateTrigger(3, TimeUnit.DAYS);
        assertThat(t.getInitialDelay(), is(0L));
        assertThat(t.getPeriod(), is(TimeUnit.DAYS.toMillis(3)));
        assertThat(t.getTimeUnit(), is(TimeUnit.DAYS));
    }

    @Test(expected = IllegalArgumentException.class)
    public void testConstructWithNegativePeriod() throws Exception {
        t = new VprUpdateTrigger(-7);
    }

    @Test
    public void testSetInitialDelay() throws Exception {
        assertThat(t.getInitialDelay(), is(TimeUnit.SECONDS.toMillis(10)));
    }

    @Test
    public void testNextExecutionTimeForInitialDelay() throws Exception {
        when(mockTriggerContext.lastScheduledExecutionTime()).thenReturn(null);

        long currentTime = System.currentTimeMillis();
        Date nextExecutionTime = t.nextExecutionTime(mockTriggerContext);
        long expectedTime = currentTime + t.getInitialDelay();
        assertThat(nextExecutionTime.getTime(), is(either(equalTo(expectedTime)).or(equalTo(expectedTime + 1L)).or(equalTo(expectedTime + 2L))));
    }

    @Test
    public void testNextExecutionTimeWithLastExecutionTimeLessThanPeriod() throws Exception {
        long lastScheduledExecutionTime = System.currentTimeMillis() - 2000L;
        long lastActualExecutionTime = lastScheduledExecutionTime + 23L;
        long lastCompletionTime = lastActualExecutionTime + 347L;
        long elapsedTime = lastCompletionTime - lastActualExecutionTime;

        when(mockTriggerContext.lastScheduledExecutionTime()).thenReturn(new Date(lastScheduledExecutionTime));
        when(mockTriggerContext.lastActualExecutionTime()).thenReturn(new Date(lastActualExecutionTime));
        when(mockTriggerContext.lastCompletionTime()).thenReturn(new Date(lastCompletionTime));

        long expectedNextExecutionTime = lastCompletionTime + (t.getPeriod() - elapsedTime);
        Date nextExecutionTime = t.nextExecutionTime(mockTriggerContext);

        assertThat(nextExecutionTime.getTime(), is(expectedNextExecutionTime));
    }

    @Test
    public void testNextExecutionTimeWithLastExecutionTimeGreaterThanPeriod() throws Exception {
        long lastScheduledExecutionTime = System.currentTimeMillis() - 0000L;
        long lastActualExecutionTime = lastScheduledExecutionTime + 42L;
        long lastCompletionTime = lastActualExecutionTime + 3000L;

        when(mockTriggerContext.lastScheduledExecutionTime()).thenReturn(new Date(lastScheduledExecutionTime));
        when(mockTriggerContext.lastActualExecutionTime()).thenReturn(new Date(lastActualExecutionTime));
        when(mockTriggerContext.lastCompletionTime()).thenReturn(new Date(lastCompletionTime));

        Date nextExecutionTime = t.nextExecutionTime(mockTriggerContext);
        assertThat(nextExecutionTime.getTime(), is(lastCompletionTime + 1));
    }
}
