package gov.va.hmp.vista.springframework.security.web;

import gov.va.hmp.vista.springframework.security.authentication.VistaAuthenticationToken;
import org.apache.commons.codec.binary.Base64;
import org.springframework.http.HttpHeaders;
import org.springframework.security.authentication.AnonymousAuthenticationToken;
import org.springframework.security.authentication.AuthenticationDetailsSource;
import org.springframework.security.authentication.AuthenticationManager;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.AuthenticationException;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.web.AuthenticationEntryPoint;
import org.springframework.security.web.authentication.NullRememberMeServices;
import org.springframework.security.web.authentication.RememberMeServices;
import org.springframework.security.web.authentication.WebAuthenticationDetailsSource;
import org.springframework.util.AntPathMatcher;
import org.springframework.util.Assert;
import org.springframework.util.PathMatcher;
import org.springframework.util.StringUtils;
import org.springframework.web.filter.GenericFilterBean;
import org.springframework.web.util.UrlPathHelper;

import javax.servlet.FilterChain;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.util.Map;

/**
 * TODO: Provide summary documentation of class VistaBasicAuthenticationFilter
 */
public class VistaBasicAuthenticationFilter extends GenericFilterBean {
    public static final String USERNAME_PASSWORD_DELIMITER = ":";
    public static final String VISTA_ID_DIVISION_DELIMITER = ";";
    public static final String ACCESS_VERIFY_CODE_DELIMITER = ";";
    public static final String URI_TEMPLATE_KEY_VISTA_ID = "vistaId";
    public static final String URI_TEMPLATE_KEY_DIVISION = "division";

    //~ Instance fields ================================================================================================

    private AuthenticationDetailsSource authenticationDetailsSource = new WebAuthenticationDetailsSource();
    private AuthenticationEntryPoint authenticationEntryPoint;
    private AuthenticationManager authenticationManager;
    private RememberMeServices rememberMeServices = new NullRememberMeServices();
    private boolean ignoreFailure = false;
    private String credentialsCharset = "UTF-8";

    private String uriTemplate = null;
    private UrlPathHelper urlPathHelper = new UrlPathHelper();
    private PathMatcher pathMatcher = new AntPathMatcher();

    //~ Methods ========================================================================================================


    public void setUriTemplate(String uriTemplate) {
        this.uriTemplate = uriTemplate;
    }

    @Override
    public void afterPropertiesSet() {
        Assert.notNull(this.authenticationManager, "An AuthenticationManager is required");

        if (!isIgnoreFailure()) {
            Assert.notNull(this.authenticationEntryPoint, "An AuthenticationEntryPoint is required");
        }
    }

    public void doFilter(ServletRequest req, ServletResponse res, FilterChain chain)
            throws IOException, ServletException {
        HttpServletRequest request = (HttpServletRequest) req;
        HttpServletResponse response = (HttpServletResponse) res;

        String header = request.getHeader(HttpHeaders.AUTHORIZATION);

        if (logger.is   ugEnabled()) {
            logger.debug("Authorization header: " + header);
        }

        if ((header != null) && header.startsWith("Basic ")) {
            byte[] base64Token = header.substring(6).getBytes("UTF-8");
            String token = new String(Base64.decodeBase64(base64Token), getCredentialsCharset(request));

            String vistaId = obtainVistaId(request, token);
            String division = obtainDivision(request, token);
            String accessCode = obtainAccessCode(request, token);
            String verifyCode = obtainVerifyCode(request, token);

            if (authenticationIsRequired(vistaId, division, accessCode, verifyCode)) {  // TODO: re-examine username strategy here - maybe use full token here
                String remoteAddress = obtainRemoteAddress(request);
                String remoteHostName = obtainRemoteHostName(request);

                VistaAuthenticationToken authRequest =
                        new VistaAuthenticationToken(vistaId, division, accessCode, verifyCode, remoteAddress, remoteHostName);
                authRequest.setDetails(authenticationDetailsSource.buildDetails(request));

                Authentication authResult;

                try {
                    authResult = authenticationManager.authenticate(authRequest);
                } catch (AuthenticationException failed) {
                    // Authentication failed
                    if (logger.is   ugEnabled()) {
                        logger.debug("Authentication request failed: " + failed.toString());
                    }

                    SecurityContextHolder.getContext().setAuthentication(null);

                    rememberMeServices.loginFail(request, response);

                    onUnsuccessfulAuthentication(request, response, failed);

                    if (ignoreFailure) {
                        chain.doFilter(request, response);
                    } else {
                        authenticationEntryPoint.commence(request, response, failed);
                    }

                    return;
                }

                // Authentication success
                if (logger.is   ugEnabled()) {
                    logger.debug("Authentication success: " + authResult.toString());
                }

                SecurityContextHolder.getContext().setAuthentication(authResult);

                rememberMeServices.loginSuccess(request, response, authResult);

                onSuccessfulAuthentication(request, response, authResult);
            }
        }

        chain.doFilter(request, response);
    }

    protected String obtainVistaId(HttpServletRequest request, String basicAuthToken) {
        String vistaId = "";
        if (StringUtils.hasText(this.uriTemplate)) {
            String path = urlPathHelper.getPathWithinApplication(request);
            if (pathMatcher.match(this.uriTemplate, path)) {
                Map<String, String> vars = pathMatcher.extractUriTemplateVariables(this.uriTemplate, path);
                if (vars.containsKey(URI_TEMPLATE_KEY_VISTA_ID)) {
                    vistaId = vars.get(URI_TEMPLATE_KEY_VISTA_ID);
                }
            }
        }
        if (StringUtils.isEmpty(vistaId)) {
            String username = obtainUsername(basicAuthToken);

            int delimiterIndex = username.indexOf(VISTA_ID_DIVISION_DELIMITER);
            if (delimiterIndex != -1) {
                vistaId = username.substring(0, delimiterIndex);
            } else {
                vistaId = username;
            }
        }
        return vistaId;
    }

    protected String obtainDivision(HttpServletRequest request, String basicAuthToken) {
        String division = "";
        String username = obtainUsername(basicAuthToken);
        int delimiterIndex = username.indexOf(VISTA_ID_DIVISION_DELIMITER);
        if (delimiterIndex != -1) {
            division = username.substring(delimiterIndex + 1);
        }
        if (StringUtils.hasText(this.uriTemplate)) {
            String path = urlPathHelper.getPathWithinApplication(request);
            if (pathMatcher.match(this.uriTemplate, path)) {
                Map<String, String> vars = pathMatcher.extractUriTemplateVariables(this.uriTemplate, path);
                if (vars.containsKey(URI_TEMPLATE_KEY_DIVISION)) {
                    division = vars.get(URI_TEMPLATE_KEY_DIVISION);
                }
            }
            if (StringUtils.isEmpty(division)) {
                division = username;
            }
        }
        return division;
    }

    protected String obtainAccessCode(HttpServletRequest request, String basicAuthToken) {
        String password = obtainPassword(basicAuthToken);
        String accessCode = "";
        int delimiterIndex = password.indexOf(ACCESS_VERIFY_CODE_DELIMITER);
        if (delimiterIndex != -1) {
            accessCode = password.substring(0, delimiterIndex);
        }
        return accessCode;
    }

    protected String obtainVerifyCode(HttpServletRequest request, String basicAuthToken) {
        String password = obtainPassword(basicAuthToken);
        String verifyCode = "";
        int delimIndex = password.indexOf(ACCESS_VERIFY_CODE_DELIMITER);
        if (delimIndex != -1) {
            verifyCode = password.substring(delimIndex + 1);
        }
        return verifyCode;
    }

    private String obtainUsername(String basicAuthToken) {
        String username = "";
        int delimIndex = basicAuthToken.indexOf(USERNAME_PASSWORD_DELIMITER);
        if (delimIndex != -1) {
            username = basicAuthToken.substring(0, delimIndex);
        }
        return username;
    }

    private String obtainPassword(String basicAuthToken) {
        String password = "";
        int delimIndex = basicAuthToken.indexOf(USERNAME_PASSWORD_DELIMITER);
        if (delimIndex != -1) {
            password = basicAuthToken.substring(delimIndex + 1);
        }
        return password;
    }

    protected String obtainRemoteAddress(HttpServletRequest request) {
        return VistaAuthenticationHttpRequestUtils.obtainRemoteAddress(request);
    }

    protected String obtainRemoteHostName(HttpServletRequest request) {
        return VistaAuthenticationHttpRequestUtils.obtainRemoteHostName(request);
    }

    private boolean authenticationIsRequired(String vistaId, String division, String accessCode, String verifyCode) {
        // Only reauthenticate if username doesn't match SecurityContextHolder and user isn't authenticated
        // (see SEC-53)
        Authentication existingAuth = SecurityContextHolder.getContext().getAuthentication();

        if (existingAuth == null || !existingAuth.isAuthenticated()) {
            return true;
        }

        // If there is an existing Vista authentication, compare relevant fields
        if (existingAuth instanceof VistaAuthenticationToken) {
            VistaAuthenticationToken vistaAuth = (VistaAuthenticationToken) existingAuth;
            if (!vistaAuth.getVistaId().equalsIgnoreCase(vistaId) &&
                    division.length() > 0 && vistaAuth.getDivision().equalsIgnoreCase(division) &&
                    !vistaAuth.getAccessCode().equalsIgnoreCase(accessCode) &&
                    !vistaAuth.getVerifyCode().equalsIgnoreCase(verifyCode)) {
                return true;
            }
        }

        // Handle unusual condition where an AnonymousAuthenticationToken is already present
        // This shouldn't happen very often, as BasicProcessingFitler is meant to be earlier in the filter
        // chain than AnonymousAuthenticationFilter. Nevertheless, presence of both an AnonymousAuthenticationToken
        // together with a BASIC authentication request header should indicate reauthentication using the
        // BASIC protocol is desirable. This behaviour is also consistent with that provided by form and digest,
        // both of which force re-authentication if the respective header is detected (and in doing so replace
        // any existing AnonymousAuthenticationToken). See SEC-610.
        if (existingAuth instanceof AnonymousAuthenticationToken) {
            return true;
        }

        return false;
    }

    protected void onSuccessfulAuthentication(HttpServletRequest request, HttpServletResponse response,
                                              Authentication authResult) throws IOException {
    }

    protected void onUnsuccessfulAuthentication(HttpServletRequest request, HttpServletResponse response,
                                                AuthenticationException failed) throws IOException {
    }

    protected AuthenticationEntryPoint getAuthenticationEntryPoint() {
        return authenticationEntryPoint;
    }

    public void setAuthenticationEntryPoint(AuthenticationEntryPoint authenticationEntryPoint) {
        this.authenticationEntryPoint = authenticationEntryPoint;
    }

    protected AuthenticationManager getAuthenticationManager() {
        return authenticationManager;
    }

    public void setAuthenticationManager(AuthenticationManager authenticationManager) {
        this.authenticationManager = authenticationManager;
    }

    protected boolean isIgnoreFailure() {
        return ignoreFailure;
    }

    public void setIgnoreFailure(boolean ignoreFailure) {
        this.ignoreFailure = ignoreFailure;
    }

    public void setAuthenticationDetailsSource(AuthenticationDetailsSource authenticationDetailsSource) {
        Assert.notNull(authenticationDetailsSource, "AuthenticationDetailsSource required");
        this.authenticationDetailsSource = authenticationDetailsSource;
    }

    public void setRememberMeServices(RememberMeServices rememberMeServices) {
        Assert.notNull(rememberMeServices, "rememberMeServices cannot be null");
        this.rememberMeServices = rememberMeServices;
    }

    public void setCredentialsCharset(String credentialsCharset) {
        Assert.hasText(credentialsCharset, "credentialsCharset cannot be null or empty");
        this.credentialsCharset = credentialsCharset;
    }

    protected String getCredentialsCharset(HttpServletRequest httpRequest) {
        return credentialsCharset;
    }
}
