package gov.va.hmp.vista.springframework.security.userdetails.rpc;

import gov.va.hmp.vista.rpc.ConnectionCallback;
import gov.va.hmp.vista.rpc.RpcHost;
import gov.va.hmp.vista.rpc.RpcOperations;
import gov.va.hmp.vista.rpc.broker.protocol.BadCredentialsException;
import gov.va.hmp.vista.rpc.broker.protocol.LockedException;
import gov.va.hmp.vista.rpc.broker.protocol.VerifyCodeExpiredException;
import gov.va.hmp.vista.rpc.conn.ConnectionUserDetails;
import gov.va.hmp.vista.springframework.security.userdetails.VistaUserDetails;
import gov.va.hmp.vista.util.RpcUriUtils;
import org.junit.Before;
import org.junit.Test;
import org.springframework.dao.PermissionDeniedDataAccessException;
import org.springframework.security.authentication.CredentialsExpiredException;

import java.util.HashMap;
import java.util.Map;

import static org.hamcrest.CoreMatchers.is;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.junit.Assert.fail;
import static org.mockito.Matchers.any;
import static org.mockito.Matchers.eq;
import static org.mockito.Mockito.*;

public class RpcTemplateUserDetailServiceTest {

    public static final String MOCK_REMOTE_ADDRESS = "123.45.67.89";
    public static final String MOCK_REMOTE_HOSTNAME = "www.example.org";

    public static final String MOCK_VISTA_ID = "9F2B";

    public static final String MOCK_PRIMARY_STATION_NUMBER = "500";
    public static final String MOCK_DIVISION = "960";
    public static final String MOCK_ACCESS_CODE = "10vehu";
    public static final String MOCK_VERIFY_CODE = "vehu10";

    public static final String MOCK_APP_HANDLE = "1A2B3C4D5E6F";

    public static final String MOCK_DUZ = "123";
    public static final Map<String,String> MOCK_DIVISION_NAMES;

    static {
        MOCK_DIVISION_NAMES = new HashMap<>();
        MOCK_DIVISION_NAMES.put(MOCK_PRIMARY_STATION_NUMBER, "CAMP MASTER");
        MOCK_DIVISION_NAMES.put(MOCK_DIVISION, "HMP DEV");
    }

    private RpcOperations mockRpcTemplate;
    private RpcTemplateUserDetailService s;

    @Before
    public void setUp() {
        mockRpcTemplate = mock(RpcOperations.class);

        s = new RpcTemplateUserDetailService();
        s.setRpcTemplate(mockRpcTemplate);
    }

    @Test
    public void testLoginWithAVCodesAndDivision() {
        ConnectionUserDetails mockConnectionUserDetails = createMockConnectionUserDetails(MOCK_DIVISION);

        RpcTemplateUserDetailService.ConnectionInfo info = new RpcTemplateUserDetailService.ConnectionInfo(new RpcHost("example.org", 0000), mockConnectionUserDetails);
        when(mockRpcTemplate.execute(any(RpcTemplateUserDetailService.ConnectionInfoCallback.class), eq("vrpcb://" + RpcUriUtils.toAuthority(MOCK_VISTA_ID, MOCK_DIVISION, MOCK_ACCESS_CODE, MOCK_VERIFY_CODE, null, null, MOCK_REMOTE_ADDRESS, MOCK_REMOTE_HOSTNAME)))).thenReturn(info);

        VistaUserDetails userDetails = s.login(MOCK_VISTA_ID, MOCK_DIVISION, MOCK_ACCESS_CODE, MOCK_VERIFY_CODE, null, null, MOCK_REMOTE_ADDRESS, MOCK_REMOTE_HOSTNAME);
        assertThat(userDetails.getDUZ(), is(MOCK_DUZ));
        assertThat(userDetails.getPrimaryStationNumber(), is(MOCK_PRIMARY_STATION_NUMBER));
        assertThat(userDetails.getDivision(), is(MOCK_DIVISION));
        assertThat(userDetails.getDivisionName(), is(MOCK_DIVISION_NAMES.get(MOCK_DIVISION)));
        assertThat(userDetails.getHost().getHostname(), is("example.org"));
        assertThat(userDetails.getHost().getPort(), is(0000));

        verify(mockRpcTemplate).execute(any(RpcTemplateUserDetailService.ConnectionInfoCallback.class), eq("vrpcb://" + RpcUriUtils.toAuthority(MOCK_VISTA_ID, MOCK_DIVISION, MOCK_ACCESS_CODE, MOCK_VERIFY_CODE, null, null, MOCK_REMOTE_ADDRESS, MOCK_REMOTE_HOSTNAME)));
    }

    private ConnectionUserDetails createMockConnectionUserDetails(String defaultDivision) {
        ConnectionUserDetails mockConnectionUserDetails = mock(ConnectionUserDetails.class);
        when(mockConnectionUserDetails.getDUZ()).thenReturn(MOCK_DUZ);
        when(mockConnectionUserDetails.getPrimaryStationNumber()).thenReturn(MOCK_PRIMARY_STATION_NUMBER);
        when(mockConnectionUserDetails.getDivision()).thenReturn(defaultDivision);
        when(mockConnectionUserDetails.getDivisionNames()).thenReturn(MOCK_DIVISION_NAMES);
        return mockConnectionUserDetails;
    }

    @Test
    public void testLoginWithAVCodesAndNoDivision() {
        ConnectionUserDetails mockConnectionUserDetails = createMockConnectionUserDetails(MOCK_PRIMARY_STATION_NUMBER);
        RpcTemplateUserDetailService.ConnectionInfo info = new RpcTemplateUserDetailService.ConnectionInfo(new RpcHost("example.org", 0000), mockConnectionUserDetails);
        when(mockRpcTemplate.execute(any(RpcTemplateUserDetailService.ConnectionInfoCallback.class), eq("vrpcb://" + RpcUriUtils.toAuthority(MOCK_VISTA_ID, null, MOCK_ACCESS_CODE, MOCK_VERIFY_CODE, null, null, MOCK_REMOTE_ADDRESS, MOCK_REMOTE_HOSTNAME)))).thenReturn(info);

        VistaUserDetails userDetails = s.login(MOCK_VISTA_ID, null, MOCK_ACCESS_CODE, MOCK_VERIFY_CODE, null, null, MOCK_REMOTE_ADDRESS, MOCK_REMOTE_HOSTNAME);
        assertThat(userDetails.getDUZ(), is(MOCK_DUZ));
        assertThat(userDetails.getPrimaryStationNumber(), is(MOCK_PRIMARY_STATION_NUMBER));
        assertThat(userDetails.getDivision(), is(MOCK_PRIMARY_STATION_NUMBER));
        assertThat(userDetails.getDivisionName(), is(MOCK_DIVISION_NAMES.get(MOCK_PRIMARY_STATION_NUMBER)));
        assertThat(userDetails.getHost().getHostname(), is("example.org"));
        assertThat(userDetails.getHost().getPort(), is(0000));

        verify(mockRpcTemplate).execute(any(RpcTemplateUserDetailService.ConnectionInfoCallback.class), eq("vrpcb://" + RpcUriUtils.toAuthority(MOCK_VISTA_ID, null, MOCK_ACCESS_CODE, MOCK_VERIFY_CODE, null, null, MOCK_REMOTE_ADDRESS, MOCK_REMOTE_HOSTNAME)));
    }

    @Test
    public void testBadCredentials() {
        when(mockRpcTemplate.execute(any(ConnectionCallback.class), eq("vrpcb://" + RpcUriUtils.toAuthority(MOCK_VISTA_ID, MOCK_DIVISION, MOCK_ACCESS_CODE, MOCK_VERIFY_CODE, MOCK_REMOTE_ADDRESS, MOCK_REMOTE_HOSTNAME)))).thenThrow(new PermissionDeniedDataAccessException("", new BadCredentialsException()));

        try {
            s.login(MOCK_VISTA_ID, MOCK_DIVISION, MOCK_ACCESS_CODE, MOCK_VERIFY_CODE, null, null, MOCK_REMOTE_ADDRESS, MOCK_REMOTE_HOSTNAME);
            fail("expected a Spring Security BadCredentialsException");
        } catch (org.springframework.security.authentication.BadCredentialsException e) {
            // NOOP
        }

        verify(mockRpcTemplate).execute(any(ConnectionCallback.class), eq("vrpcb://" + RpcUriUtils.toAuthority(MOCK_VISTA_ID, MOCK_DIVISION, MOCK_ACCESS_CODE, MOCK_VERIFY_CODE, MOCK_REMOTE_ADDRESS, MOCK_REMOTE_HOSTNAME)));
    }

    @Test
    public void testVerifyCodeExpired() {
        when(mockRpcTemplate.execute(any(ConnectionCallback.class), eq("vrpcb://" + RpcUriUtils.toAuthority(MOCK_VISTA_ID, MOCK_DIVISION, MOCK_ACCESS_CODE, MOCK_VERIFY_CODE, MOCK_REMOTE_ADDRESS, MOCK_REMOTE_HOSTNAME)))).thenThrow(new PermissionDeniedDataAccessException("",new VerifyCodeExpiredException()));

        try {
            s.login(MOCK_VISTA_ID, MOCK_DIVISION, MOCK_ACCESS_CODE, MOCK_VERIFY_CODE, null, null, MOCK_REMOTE_ADDRESS, MOCK_REMOTE_HOSTNAME);
            fail("expected a Spring Security BadCredentialsException");
        } catch (CredentialsExpiredException e) {
            // NOOP
        }

        verify(mockRpcTemplate).execute(any(ConnectionCallback.class), eq("vrpcb://" + RpcUriUtils.toAuthority(MOCK_VISTA_ID, MOCK_DIVISION, MOCK_ACCESS_CODE, MOCK_VERIFY_CODE, MOCK_REMOTE_ADDRESS, MOCK_REMOTE_HOSTNAME)));
    }

    @Test
    public void testLoginWithAppHandleAndNoDivision() {
        ConnectionUserDetails mockConnectionUserDetails = createMockConnectionUserDetails(MOCK_PRIMARY_STATION_NUMBER);
        RpcTemplateUserDetailService.ConnectionInfo info = new RpcTemplateUserDetailService.ConnectionInfo(new RpcHost("example.org", 0000), mockConnectionUserDetails);
        when(mockRpcTemplate.execute(any(RpcTemplateUserDetailService.ConnectionInfoCallback.class), eq("vrpcb://" + RpcUriUtils.toAuthority(MOCK_VISTA_ID, null, MOCK_APP_HANDLE, MOCK_REMOTE_ADDRESS, MOCK_REMOTE_HOSTNAME)))).thenReturn(info);

        s.login(MOCK_VISTA_ID, null, MOCK_APP_HANDLE, MOCK_REMOTE_ADDRESS, MOCK_REMOTE_HOSTNAME);

        verify(mockRpcTemplate).execute(any(RpcTemplateUserDetailService.ConnectionInfoCallback.class), eq("vrpcb://" + RpcUriUtils.toAuthority(MOCK_VISTA_ID, null, MOCK_APP_HANDLE, MOCK_REMOTE_ADDRESS, MOCK_REMOTE_HOSTNAME)));
    }

    @Test
    public void testLocked() {
        when(mockRpcTemplate.execute(any(ConnectionCallback.class), eq("vrpcb://" + RpcUriUtils.toAuthority(MOCK_VISTA_ID, MOCK_DIVISION, MOCK_ACCESS_CODE, MOCK_VERIFY_CODE, MOCK_REMOTE_ADDRESS, MOCK_REMOTE_HOSTNAME)))).thenThrow(new PermissionDeniedDataAccessException("", new LockedException("sfdsfwesdfsdf")));

        try {
            s.login(MOCK_VISTA_ID, MOCK_DIVISION, MOCK_ACCESS_CODE, MOCK_VERIFY_CODE, null, null, MOCK_REMOTE_ADDRESS, MOCK_REMOTE_HOSTNAME);
            fail("expected a Spring Security BadCredentialsException");
        } catch (org.springframework.security.authentication.LockedException e) {
            // NOOP
        }

        verify(mockRpcTemplate).execute(any(ConnectionCallback.class), eq("vrpcb://" + RpcUriUtils.toAuthority(MOCK_VISTA_ID, MOCK_DIVISION, MOCK_ACCESS_CODE, MOCK_VERIFY_CODE, MOCK_REMOTE_ADDRESS, MOCK_REMOTE_HOSTNAME)));
    }
}
