# VistA RPC Client

The VistA Remote Procedure Call (RPC) client is a bridge connecting java applications to M-based data and business rules
on VistA systems through the execution of VistA Remote Procedure Calls (RPCs).  It does this using the same network
protocol as the mature [VistA RPC Broker][], in use by the [Computerized Patient Record System (CPRS)][]
and many other clinical applications in the VA, but with a streamlined and easy to use API.

## VistA RPC URIs

The HMP VistA RPC Client introduces the concept of expressing to VistA RPCs via a [URI][]. This is a slightly different model
than the stateful connection employed by the traditional [RPC Broker][VistA RPC Broker], but is very powerful and more
compatible with web technologies.  In particular, they are proving to be very helpful in managing the logistics of
calling RPCs in a patient-centric rather than a facility-centric manner. Because all of the connection information for
a VistA RPC is contained in one string, we can refer to many VistA systems and their RPCs relatively easily.  Thinking
of executing an RPC as a logically stateless operation (like fetching a resource at an `http://` URL) makes developing
against this API much eaiser in our experience; for example, calling an RPC from a java program can be done in one line.

VistA RPC URIs are of this form:
### VistA RPC URI Format
```
{protocol}://{access-code};{verify-code}@{host}:{port}/{rpc-context}/{rpc-name}
```

Today, there is only one protocol supported by the client, the `vrpcb` protocol, which is the protocol used by the
[RPC Broker][VistA RPC Broker], so all the RPC URIs look like this:
#### VistA RPC URI Format for RPC Broker Protocol
```
vrpcb://{access-code};{verify-code}@{host}:{port}/{rpc-context}/{rpc-name}
```

Hypothetically, we can specify a URI for the VistALink protocol like so:
#### VistA RPC URI Format for VistALink Protocol (Hypothetical)
```
vlink://{access-code};{verify-code}@{host}:{port}/{rpc-context}/{rpc-name}
```

Some components of the URI are optional and/or implicit based on how you have configured the client.  More detail on
this below.

#### Multidivisional VistA Systems
Some VistA systems are configured so that they host the applications and data for multiple facilities.  In these
multidivisional systems, it is sometimes necessary to execute RPCs in a division other than the default to get the
correct behavior.  The optionally specified division goes in the URI like this:
```
vrpcb://{division}:{access-code};{verify-code}@{host}:{port}/{rpc-context}/{rpc-name}
```

### Example URIs
The URI for the CPRS RPC that fetches the list of vitals for the coversheet from a VistA system located at `example.org`
with RPC Broker listener listening on port `0000` executing as the `VEHU,TEN` user is:
```
vrpcb://10vehu;vehu10@example.org:0000/OR CPRS GUI CHART/ORQQVI VITALS
```

Or to use the VPR data extraction RPC:
```
vrpcb://10vehu;vehu10@example.org:0000/VPR SYNCHRONIZATION CONTEXT/GET VISTA DATA
```

Or to use the VPR data extraction RPC in division `500D` instead of the default division:
```
vrpcb://500D:10vehu;vehu10@example.org:0000/VPR SYNCHRONIZATION CONTEXT/GET VISTA DATA
```

The URI for fetching the VistA system's welcome message (`rpc-context` is not required and is omitted) is:
```
vrpcb://10vehu;vehu10@example.org:0000/XUS INTRO MSG
```

### URIs without credentials
It is possible to configure the RPC client so that credentials are added programmatically so they don't need to be
included in the URI.  The URI would like this, in that case:
```
vrpcb://example.org:0000/OR CPRS GUI CHART/ORQQVI VITALS
```

Or, in division `500D`, instead of the default division:
```
vrpcb://500D@example.org:0000/OR CPRS GUI CHART/ORQQVI VITALS
```

More information about configuring the RPC client like this is included below under [CredentialsProvider](#credentialsProvider).

### URIs without host and port
In addition to leaving out credentials, it is possible to configure the RPC client so that connection info are added
programmatically and don't need to be included in the URI.  The URI would like this, in that case:
```
vrpcb:///OR CPRS GUI CHART/ORQQVI VITALS
```

or, defaulting the protocol, the relative URI:
```
/OR CPRS GUI CHART/ORQQVI VITALS
```

More information about configuring the RPC client so that these forms of URI execute are included below under [RpcHostResolver](#rpcHostResolver).

## RpcTemplate

The `RpcTemplate` class is the central class in the VistA RPC Client. It follows the Spring Framework data access idioms
and emulates the  `RestTemplate` and `JbdcTemplate` classes in particular.

The RpcTemplate class simplifies the use of VistA RPCs and helps to avoid common errors. It executes core broker workflow,
leaving application code to call RPCs and extract results. This class executes RPCs, optionally initiating iteration
over the result lines and catching broker exceptions and translating them to the generic, more informative exception
hierarchy defined in the `org.springframework.dao` package.

Instances of `java.util.List`, `java.util.Map`, and `com.fasterxml.jackson.databind.JsonNode` passed as
parameters to the RPC via any of the execute*() methods are translated into multiples for transfer over the wire.  All
other objects passed as parameters via execute*() are passed to VistA as Strings by calling their toString() method.
There is a 240 character limit inherent in the RPC Broker network protocol for RPC parameters (of type `LITERAL`) passed
in this way.

## Samples

```java
import gov.va.cpe.vista.rpc.RpcTemplate;

public class VistARPCRunner {
    public static void main(String[] args) {
        try {
            RpcTemplate rpcTemplate = new RpcTemplate();

            System.out.println(rpcTemplate.executeForString("vrpcb://10vehu;vehu10@example.org:0000/OR CPRS GUI CHART/ORQQVI VITALS", "229"));
        } finally {
            rpcTemplate.dispose();
        }
    }
}
```

#### Execute an RPC returning the response as an int
```java
int result = rpcTemplate.executeForInt("vrpcb://10vehu;vehu10@example.org:0000/OR CPRS GUI CHART/ORQQVI VITALS", "229");
```

#### Execute an RPC returning the result as an array of String, one per line of the response
```java
String[] result = rpcTemplate.executeForLines("vrpcb://10vehu;vehu10@example.org:0000/OR CPRS GUI CHART/ORQQVI VITALS", "229");
```

#### Execute an RPC returning the result as a List, one per line of the response
```java
List<String> result = rpcTemplate.executeForList(String.class, "vrpcb://10vehu;vehu10@example.org:0000/OR CPRS GUI CHART/ORQQVI VITALS", "229");
```

#### Execute an RPC returning the result as a List, populating a number of objects
```java
LineMapper<VitalSign> lineMapper = new LineMapper<VitalSign>() {
    VitalSign mapLine(String line, int lineNum) {
    VitalSign v = new VitalSign();
    v.setName(VistaStringUtils.piece(line, "^", 2));
    v.setValue(VistaStringUtils.piece(line, "^", 3));
    return v;
    }
};

List<VitalSign> vitals = rpcTemplate.executeForList(lineMapper, "vrpcb://10vehu;vehu10@example.org:0000/OR CPRS GUI CHART/ORQQVI VITALS", "229");
```

#### Execute an RPC returning the result as a single object
```java
RpcResponseExtractor<VitalSignSet> responseExtractor = new RpcResponseExtractor<VitalSignSet>() {
    VitalSignSet extractData(RpcResponse response) {
        VitalSignSet v = new VitalSignSet();
        // fill in the vital signs here from lines of the response
        return v;
    }
};

VitalSignSet vitals = rpcTemplate.executeForObject(responseExtractor, "vrpcb://10vehu;vehu10@example.org:0000/OR CPRS GUI CHART/ORQQVI VITALS", "229");
```

### CredentialsProvider

In order to avoid having to include user credentials in every URI (to facilitate calling RPCs on behalf of a signed
in user, for example) RpcTemplate can be configured to use an instance of a `CredentialsProvider` to supply user
credentials for RPC URIs for which they are missing.

For example, `vrpcb://example.org:0000/FOO/BAR` is valid but not sufficient for executing the `BAR` remote procedure
call in the `FOO` context at `example.org` port `0000`. `RpcTemplate` delegates the addition of replacement of the
`userInfo` part of the `vrpcb://{userInfo}@example.org:0000/FOO/BAR` URI to implementations of the
`CredentialsProvider` interface.  Below is an example where one set of credentials are used for all RPCs:

```java
import gov.va.cpe.vista.rpc.RpcTemplate;
import gov.va.cpe.vista.rpc.CredentialsProvider;
import gov.va.cpe.vista.rpc.RpcHost;

public class VistARPCRunner {
    public static void main(String[] args) {
        try {
            RpcTemplate rpcTemplate = new RpcTemplate();
            rpcTemplate.setCredentialProvider(new CredentialsProvider() {
                public String getCredentials(RpcHost host, String userInfo) {
                    return "10vehu;vehu10";
                }
            });

            System.out.println(rpcTemplate.executeForString("vrpcb://example.org:0000/OR CPRS GUI CHART/ORQQVI VITALS", "229"));
        } finally {
            rpcTemplate.dispose();
        }
    }
}
```

### RpcHostResolver

In order to avoid having to include the host and port of a VistA server in every URI, `RpcTemplate` can be configured to
use an instance of a `RpcHostResolver` to resolve ambiguous or encoded hostnames into RpcHost instances that contain
the required VistA system connection information.

For example, `vrpcb://ABCD/FOO/BAR` is valid but not sufficient for executing the `BAR` remote procedure call in
the `FOO` context.  `RpcTemplate` delegates the resolution of `ABCD` into a hostname and port (an instance
of `RpcHost` so the RPC can be executed in the right VistA system. Below is an example of resolving ambiguous host
aliases to a small set of predefined systems:

```java
import gov.va.cpe.vista.rpc.RpcTemplate;
import gov.va.cpe.vista.rpc.RpcHost;
import gov.va.hmp.vista.rpc.RpcHostResolver;
import gov.va.hmp.vista.rpc.broker.conn.VistaIdNotFoundException;

public class VistARPCRunner {

    private static Map<String,RpcHost> hosts = new HashMap<>();

    static {
        hosts.put("VISTA_A", new RpcHost(a.example.org, 0000));
        hosts.put("VISTA_B", new RpcHost(b.example.org, 9061));
    }

    public static void main(String[] args) {
        try {
            RpcTemplate rpcTemplate = new RpcTemplate();
            rpcTemplate.setHostResolver(new RpcHostResolver() {
                public RpcHost resolve(String vistaId) throws VistaIdNotFoundException {
                    if (!hosts.containsKey(vistaId)) throw new VistaIdNotFoundException(vistaId);
                    return hosts.get(vistaId);
                }
            });

            System.out.println(rpcTemplate.executeForString("vrpcb://10vehu;vehu10@VISTA_A/OR CPRS GUI CHART/ORQQVI VITALS", "229"));
            System.out.println(rpcTemplate.executeForString("vrpcb://10vehu;vehu10@VISTA_B/OR CPRS GUI CHART/ORQQVI VITALS", "229"));
        } finally {
            rpcTemplate.dispose();
        }
    }
}
```

[Computerized Patient Record System (CPRS)]: http://www.domain/vdl/application.asp?appid=61
[URI]: http://en.wikipedia.org/wiki/Uniform_resource_identifier
[VistA RPC Broker]: http://www.domain/vdl/application.asp?appid=23
[VistALink]: http://www.domain/vdl/application.asp?appid=163
