package gov.va.hmp.vista.rpc.util;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ArrayNode;

import gov.va.hmp.vista.rpc.*;
import gov.va.hmp.vista.rpc.conn.SystemInfo;
import gov.va.hmp.vista.util.RpcUriUtils;
import joptsimple.OptionException;
import joptsimple.OptionParser;
import joptsimple.OptionSet;
import joptsimple.OptionSpec;

import org.apache.commons.codec.binary.Base64;
import org.springframework.dao.DataAccessException;
import org.springframework.web.util.UriComponents;
import org.springframework.web.util.UriComponentsBuilder;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.spec.SecretKeySpec;

import java.io.BufferedReader;
import java.io.Console;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.security.InvalidKeyException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

import static gov.va.hmp.vista.util.RpcUriUtils.VISTA_RPC_BROKER_SCHEME;

public class Main {
    public static void main(String[] args) throws Exception {
        OptionParser parser = new OptionParser();
        OptionSpec<String> pathToVistaAccountsFile = parser.acceptsAll(Arrays.asList("f", "F", "file"), "Path to HMP vista-accounts.json file").withRequiredArg().ofType(String.class);
        OptionSpec<String> hmpServerId = parser.acceptsAll(Arrays.asList("m", "M", "hmpServerId"), "HMP Server ID").requiredIf(pathToVistaAccountsFile).withRequiredArg().ofType(String.class);
        OptionSpec<String> server = parser.acceptsAll(Arrays.asList("s", "S", "server", "host"), "VistA Server").requiredUnless(pathToVistaAccountsFile).withRequiredArg().ofType(String.class);
        OptionSpec<Integer> port = parser.acceptsAll(Arrays.asList("p", "P", "port"), "RPC Broker Port").withRequiredArg().ofType(Integer.class).defaultsTo(RpcUriUtils.DEFAULT_PORT);
        OptionSpec<String> rpcContext = parser.acceptsAll(Arrays.asList("c", "context"), "RPC Context").withRequiredArg().ofType(String.class);
        OptionSpec<String> rpcName = parser.acceptsAll(Arrays.asList("r", "rpc"), "RPC Name").withRequiredArg().ofType(String.class);
        OptionSpec<String> rpcArgs = parser.acceptsAll(Arrays.asList("a", "arg"), "RPC Argument <key=value>").withRequiredArg().ofType(String.class);
        OptionSpec<Void> verbose = parser.acceptsAll(Arrays.asList("v", "verbose"), "Verbose output");
        OptionSpec<Void> help = parser.acceptsAll(Arrays.asList("h", "?", "help"), "Show Help").forHelp();
        
        try {
            OptionSet options = parser.parse(args);
            
            // get params/args (if any)
            Map<String, String> rpcArgVals = new HashMap<>();
            for (String key : options.valuesOf(rpcArgs)) {
            	String[] split = key.split("=");
            	if (split.length == 2) {
            		rpcArgVals.put(split[0], split[1]);
            	}
            }
            if (!rpcArgVals.isEmpty() && options.has(verbose)) {
            	System.out.println("RPC Args: " + rpcArgVals);
            }

            if (options.has(pathToVistaAccountsFile) && options.has(hmpServerId)) {
                ObjectMapper mapper = new ObjectMapper();
                JsonNode vaNode = mapper.readTree(new FileInputStream(options.valueOf(pathToVistaAccountsFile)));
                ArrayNode accountsNode = (ArrayNode) vaNode.path("data").path("items");
                for (JsonNode acct : accountsNode) {
                    VistaAccount vacnt = new VistaAccount(acct, options.valueOf(hmpServerId));
                    if (options.has(rpcContext) && options.has(rpcName)) {
                        testHMPVistaAccount(vacnt, options.valueOf(rpcContext), options.valueOf(rpcName), rpcArgVals, options.has(verbose));
                    } else {
                        testHMPVistaAccount(vacnt, "VPR SYNCHRONIZATION CONTEXT", "VPR DATA VERSION", null, options.has(verbose));
                    }
                }
            } else if (options.has(server) && options.has(rpcContext) && options.has(rpcName)) {
                Console console = System.console();
                char[] accessCode = console.readPassword("Access Code: ");
                char[] verifyCode = console.readPassword("Verify Code: ");
                UriComponents uri = UriComponentsBuilder.newInstance()
                        .scheme(VISTA_RPC_BROKER_SCHEME)
                        .userInfo(new String(accessCode) + ";" + new String(verifyCode))
                        .host(options.valueOf(server))
                        .port(options.valueOf(port))
                        .pathSegment(options.valueOf(rpcContext), options.valueOf(rpcName)).build();
                
                RpcTemplate t = new RpcTemplate();
                try {
                    RpcResponse r = rpcArgVals != null && !rpcArgVals.isEmpty() ? t.execute(uri.toUriString(), rpcArgVals) : t.execute(uri.toUriString());
                    if (options.has(verbose)) {
                        echoVerbose(r);
                    }
                    System.out.println(r.toString());
                } catch (DataAccessException e) {
                    System.out.println("CONNECTION FAILURE: " + e.getCause().getMessage());
                } finally {
                    t.destroy();
                }
            } else if (options.has(server)) {
                RpcHost host = new RpcHost(options.valueOf(server), options.valueOf(port));

                RpcTemplate t = new RpcTemplate();
                try {
                    SystemInfo info = t.fetchSystemInfo(host);

                    if (options.has(verbose)) {
                    	System.out.println("CONNECTION SUCCESSFUL: HMP VistA ID='" + info.getVistaId() + "', intro message below");
                    }
                    System.out.println(info.getIntroText());
                } catch (DataAccessException e) {
                    System.out.println("CONNECTION FAILURE: " + e.getCause().getMessage());
                } finally {
                    t.destroy();
                }
            } else {
                parser.printHelpOn(System.out);
            }
        } catch (OptionException e) {
            System.out.println(e.getMessage());
            parser.printHelpOn(System.out);
        }
    }

    private static void testHMPVistaAccount(VistaAccount vac, String rpcContext, String rpcName, Map<String, String> rpcArgs, boolean verbose) throws Exception {
        if (verbose) System.out.println("Testing account: " + vac.name);
        UriComponents uri = UriComponentsBuilder.newInstance()
                .scheme(VISTA_RPC_BROKER_SCHEME)
                .userInfo(vac.accessCode + ";" + vac.verifyCode)
                .host(vac.host)
                .port(vac.port)
                .pathSegment(rpcContext, rpcName).build();

        RpcTemplate t = new RpcTemplate();
        try {
            RpcResponse r = rpcArgs != null && !rpcArgs.isEmpty() ? t.execute(uri.toUriString(), rpcArgs) : t.execute(uri.toUriString());

            if (verbose) {
                echoVerbose(r);
            }
            System.out.println(r.toString());
        } catch (DataAccessException e) {
            System.out.println("CONNECTION FAILURE: " + e.getCause().getMessage());
        } finally {
            t.destroy();
        }
    }

    private static void echoVerbose(RpcResponse r) {
        System.out.println("REQUEST URI: " + r.getRequestUri());
        System.out.println("CONNECTION SUCCESSFUL: HMP VistA ID='" + r.getVistaId() + "', DIVISION='" + r.getDivision() + "', DUZ='" + r.getDUZ() + "', elapsed millis=" + r.getElapsedMillis());
    }

    private static class VistaAccount {
        private String vistaId;
        private String name;
        private String host;
        private Integer port;
        private String accessCode;
        private String verifyCode;

        private VistaAccount(String vistaId, String name, String host, Integer port, String credentials, Boolean credsEncrypted, String hmpServerId) {
            this.vistaId = vistaId;
            this.name = name;
            this.host = host;
            this.port = port;
            if (credsEncrypted) {
                HmpEncryption enc = HmpEncryption.getInstance(hmpServerId, this.vistaId);
                try {
                    credentials = enc.decrypt(credentials, hmpServerId, vistaId);
                } catch (Exception e) {
                    e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                }
            }
            accessCode = credentials.split(";")[0];
            verifyCode = credentials.split(";")[1];
        }

        private VistaAccount(JsonNode node, String hmpServerId) {
            this.vistaId = node.path("vistaId").textValue();
            this.name = node.path("name").textValue();
            this.host = node.path("host").textValue();
            this.port = node.path("port").intValue();
            String credentials = node.path("vprUserCredentials").textValue();
            Boolean encrypted = node.path("encrypted").booleanValue();
            if (encrypted) {
                HmpEncryption enc = HmpEncryption.getInstance(hmpServerId, this.vistaId);
                try {
                    credentials = enc.decrypt(credentials, hmpServerId, vistaId);
                } catch (Exception e) {
                    e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                }
            }
            accessCode = credentials.split(";")[0];
            verifyCode = credentials.split(";")[1];
        }
    }

    private static class HmpEncryption {

        private boolean cinit = false;
        private Cipher cipher;
        private Cipher decipher;
        private Cipher encipher;

        public String decrypt(String str, String hmpServerId, String vistaSystemId) throws BadPaddingException, IllegalBlockSizeException, NoSuchPaddingException, NoSuchAlgorithmException, InvalidKeyException {
            if (!cinit) {
                cinit(hmpServerId, vistaSystemId);
            }
            byte[] encryptedBytes = Base64.decodeBase64(str);
            String decrypted = new String(decipher.doFinal(encryptedBytes));
            return decrypted;
        }

        public String encrypt(String str, String hmpServerId, String vistaSystemId) throws BadPaddingException, IllegalBlockSizeException, NoSuchPaddingException, NoSuchAlgorithmException, InvalidKeyException {
            if (!cinit) {
                cinit(hmpServerId, vistaSystemId);
            }
            String encrypted = new String(Base64.encodeBase64(encipher.doFinal(str.getBytes())));
            return encrypted;
        }

        private byte[] salt;
        private SecretKeySpec dodgeball;

        private synchronized void cinit(String hmpServerId, String vistaSystemId) throws NoSuchPaddingException, NoSuchAlgorithmException, InvalidKeyException {
            if (dodgeball == null || salt == null) {
                MessageDigest digest = MessageDigest.getInstance("SHA");
                String svr = hmpServerId;
                byte[] sbites = svr.getBytes();
                digest.update(sbites);
                dodgeball = new SecretKeySpec(digest.digest(), 0, 16, "AES");
                salt = vistaSystemId.getBytes();
            }
            decipher = Cipher.getInstance("AES/ECB/PKCS5Padding");
            decipher.init(Cipher.DECRYPT_MODE, dodgeball);
            encipher = Cipher.getInstance("AES/ECB/PKCS5Padding");
            encipher.init(Cipher.ENCRYPT_MODE, dodgeball);
            cinit = true;
        }

        public static HmpEncryption getInstance(String hmpServerId, String vistaSystemId) {
            HmpEncryption venc = new HmpEncryption();
            MessageDigest digest = null;
            try {
                digest = MessageDigest.getInstance("SHA");
                byte[] pbites = hmpServerId.getBytes();
                digest.update(pbites);
                venc.dodgeball = new SecretKeySpec(digest.digest(), 0, 16, "AES");
                venc.salt = vistaSystemId.getBytes();
            } catch (NoSuchAlgorithmException e) {
                e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
            }

            return venc;
        }
    }
}
